# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An online Q-learning agent with PopArt on Catch with large rewards.

Learns faster with PopArt than without.
"""

import collections
from absl import app
from absl import flags
from bsuite.environments import catch
from bsuite.utils import wrappers
import haiku as hk
from haiku import nets
import jax
import jax.numpy as jnp
import optax
import rlax
from rlax.examples import experiment

ActorOutput = collections.namedtuple("ActorOutput", "actions")
Transition = collections.namedtuple("Transition",
                                    "obs_tm1 a_tm1 r_t discount_t obs_t")

FLAGS = flags.FLAGS
flags.DEFINE_integer("seed", 42, "Random seed.")
flags.DEFINE_float("reward_scale", 10000, "Reward scale on Catch.")
flags.DEFINE_integer("train_episodes", 1000, "Number of train episodes.")
flags.DEFINE_integer("num_hidden_units", 50, "Number of network hidden units.")
flags.DEFINE_float("epsilon", 0.01, "Epsilon-greedy exploration probability.")
flags.DEFINE_float("discount_factor", 0.99, "Q-learning discount factor.")
flags.DEFINE_float("learning_rate", 0.005, "Optimizer learning rate.")
flags.DEFINE_float("pop_art_step_size", 3e-3, "PopArt normalization step size.")
flags.DEFINE_integer("eval_episodes", 100, "Number of evaluation episodes.")
flags.DEFINE_integer("evaluate_every", 50,
                     "Number of episodes between evaluations.")


def build_network(num_hidden_units: int, num_actions: int) -> hk.Transformed:
  """Factory for a simple MLP network for approximating Q-values."""

  def q(obs):
    flatten = lambda x: jnp.reshape(x, (-1,))
    network = hk.Sequential(
        [flatten, nets.MLP([num_hidden_units, num_actions])])
    return network(obs)

  return hk.without_apply_rng(hk.transform(q))


class TransitionAccumulator:
  """Simple Python accumulator for transitions."""

  def __init__(self):
    self._prev = None
    self._action = None
    self._latest = None

  def push(self, env_output, action):
    self._prev = self._latest
    self._action = action
    self._latest = env_output

  def sample(self, batch_size):
    assert batch_size == 1
    return Transition(self._prev.observation, self._action, self._latest.reward,
                      self._latest.discount, self._latest.observation)

  def is_ready(self, batch_size):
    assert batch_size == 1
    return self._prev is not None


class PopArtAgent:
  """An online Q-learning deep RL agent with PopArt."""

  def __init__(self, observation_spec, action_spec, num_hidden_units, epsilon,
               learning_rate, pop_art_step_size):
    self._observation_spec = observation_spec
    self._action_spec = action_spec
    self._epsilon = epsilon
    # Neural net and optimiser.
    self._network = build_network(num_hidden_units, action_spec.num_values)
    self._optimizer = optax.adam(learning_rate)
    # Jitting for speed.
    self.actor_step = jax.jit(self.actor_step)
    self.learner_step = jax.jit(self.learner_step)
    self._initial_pop_art_state, self._pop_art_update = rlax.popart(
        num_outputs=1, step_size=pop_art_step_size, scale_lb=1e-5, scale_ub=1e5)

  def initial_params(self, key):
    sample_input = self._observation_spec.generate_value()
    return self._network.init(key, sample_input)

  def initial_actor_state(self):
    return ()

  def initial_learner_state(self, params):
    return self._optimizer.init(params), self._initial_pop_art_state()

  def actor_step(self, params, env_output, actor_state, key, evaluation):
    norm_q = self._network.apply(params, env_output.observation)
    # This is equivalent to epsilon-greedy on the (unnormalized) Q-values
    # because normalization is linear, therefore the argmaxes are the same.
    train_a = rlax.epsilon_greedy(self._epsilon).sample(key, norm_q)
    eval_a = rlax.greedy().sample(key, norm_q)
    a = jax.lax.select(evaluation, eval_a, train_a)
    return ActorOutput(actions=a), actor_state

  def learner_step(self, params, data, learner_state, unused_key):
    opt_state, pop_art_state = learner_state
    dloss_dtheta, pop_art_state = jax.grad(
        self._loss, has_aux=True)(params, pop_art_state, *data)
    updates, opt_state = self._optimizer.update(dloss_dtheta, opt_state)
    params = optax.apply_updates(params, updates)
    return params, (opt_state, pop_art_state)

  def _loss(self, params, pop_art_state, obs_tm1, a_tm1, r_t, discount_t,
            obs_t):
    """Loss function."""
    indices = jnp.array(0)  # Only one output for normalization.

    # Calculate targets by unnormalizing Q-values output by network.
    norm_q_t = self._network.apply(params, obs_t)
    q_t = rlax.unnormalize(pop_art_state, norm_q_t, indices)
    target_tm1 = r_t + discount_t * jnp.max(q_t)

    # Update PopArt statistics and use them to update the network parameters to
    # POP (preserve outputs precisely). If there were target networks, the
    # parameters for these would also need to be updated.
    final_linear_module_name = "mlp/~/linear_1"
    mutable_params = hk.data_structures.to_mutable_dict(params)
    linear_params = mutable_params[final_linear_module_name]
    popped_linear_params, new_pop_art_state = self._pop_art_update(
        params=linear_params, state=pop_art_state, targets=target_tm1,
        indices=indices)
    mutable_params[final_linear_module_name] = popped_linear_params
    popped_params = hk.data_structures.to_immutable_dict(mutable_params)

    # Normalize target with updated PopArt statistics.
    norm_target_tm1 = rlax.normalize(new_pop_art_state, target_tm1, indices)

    # Calculate parameter update with normalized target and popped parameters.
    norm_q_t = self._network.apply(popped_params, obs_t)
    norm_q_tm1 = self._network.apply(popped_params, obs_tm1)
    td_error = jax.lax.stop_gradient(norm_target_tm1) - norm_q_tm1[a_tm1]
    return rlax.l2_loss(td_error), new_pop_art_state


def main(unused_arg):
  env = catch.Catch(seed=FLAGS.seed)
  env = wrappers.RewardScale(env, reward_scale=FLAGS.reward_scale)
  agent = PopArtAgent(
      observation_spec=env.observation_spec(),
      action_spec=env.action_spec(),
      num_hidden_units=FLAGS.num_hidden_units,
      epsilon=FLAGS.epsilon,
      learning_rate=FLAGS.learning_rate,
      pop_art_step_size=FLAGS.pop_art_step_size,
  )

  accumulator = TransitionAccumulator()
  experiment.run_loop(
      agent=agent,
      environment=env,
      accumulator=accumulator,
      seed=FLAGS.seed,
      batch_size=1,
      train_episodes=FLAGS.train_episodes,
      evaluate_every=FLAGS.evaluate_every,
      eval_episodes=FLAGS.eval_episodes,
  )


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions implementing custom non-linear transformations.

This is a collection of element-wise non-linear transformations that may be
used to transform losses, value estimates, or other multidimensional data.
"""

import chex
import jax
import jax.numpy as jnp
from rlax._src import base

Array = chex.Array
Numeric = chex.Numeric


def identity(x: Array) -> Array:
  """Identity transform."""
  chex.assert_type(x, float)
  return x


def sigmoid(x: Numeric) -> Array:
  """Sigmoid transform."""
  chex.assert_type(x, float)
  return jax.nn.sigmoid(x)


def logit(x: Array) -> Array:
  """Logit transform, inverse of sigmoid."""
  chex.assert_type(x, float)
  return -jnp.log(1. / x - 1.)


def signed_logp1(x: Array) -> Array:
  """Signed logarithm of x + 1."""
  chex.assert_type(x, float)
  return jnp.sign(x) * jnp.log1p(jnp.abs(x))


def signed_expm1(x: Array) -> Array:
  """Signed exponential of x - 1, inverse of signed_logp1."""
  chex.assert_type(x, float)
  return jnp.sign(x) * jnp.expm1(jnp.abs(x))


def signed_hyperbolic(x: Array, eps: float = 1e-3) -> Array:
  """Signed hyperbolic transform, inverse of signed_parabolic."""
  chex.assert_type(x, float)
  return jnp.sign(x) * (jnp.sqrt(jnp.abs(x) + 1) - 1) + eps * x


def hyperbolic_sin(x: Array) -> Array:
  """Hyperbolic sinus transform."""
  chex.assert_type(x, float)
  return jnp.sinh(x)


def hyperbolic_arcsin(x: Array) -> Array:
  """Hyperbolic arcsinus transform."""
  chex.assert_type(x, float)
  return jnp.arcsinh(x)


def signed_parabolic(x: Array, eps: float = 1e-3) -> Array:
  """Signed parabolic transform, inverse of signed_hyperbolic."""
  chex.assert_type(x, float)
  z = jnp.sqrt(1 + 4 * eps * (eps + 1 + jnp.abs(x))) / 2 / eps - 1 / 2 / eps
  return jnp.sign(x) * (jnp.square(z) - 1)


def power(x: Array, p: float) -> Array:
  """Power transform; `power_tx(_, 1/p)` is the inverse of `power_tx(_, p)`."""
  chex.assert_type(x, float)
  q = jnp.sqrt(p)
  return jnp.sign(x) * (jnp.power(jnp.abs(x) / q + 1., p) - 1) / q


def transform_to_2hot(
    scalar: Array,
    min_value: float,
    max_value: float,
    num_bins: int) -> Array:
  """Transforms a scalar tensor to a 2 hot representation."""
  scalar = jnp.clip(scalar, min_value, max_value)
  scalar_bin = (scalar - min_value) / (max_value - min_value) * (num_bins - 1)
  lower, upper = jnp.floor(scalar_bin), jnp.ceil(scalar_bin)
  lower_value = (lower / (num_bins - 1.0)) * (max_value - min_value) + min_value
  upper_value = (upper / (num_bins - 1.0)) * (max_value - min_value) + min_value
  p_lower = (upper_value - scalar) / (upper_value - lower_value + 1e-5)
  p_upper = 1 - p_lower
  lower_one_hot = base.one_hot(
      lower, num_bins, dtype=scalar.dtype) * jnp.expand_dims(p_lower, -1)
  upper_one_hot = base.one_hot(
      upper, num_bins, dtype=scalar.dtype) * jnp.expand_dims(p_upper, -1)
  return lower_one_hot + upper_one_hot


def transform_from_2hot(
    probs: Array,
    min_value: float,
    max_value: float,
    num_bins: int) -> Array:
  """Transforms from a categorical distribution to a scalar."""
  support_space = jnp.linspace(
      min_value, max_value, num_bins).astype(probs.dtype)
  scalar = jnp.sum(probs * jnp.expand_dims(support_space, 0), -1)
  return scalar


def transform_to_2hot_nonlinear(scalar: Array, bins: Array) -> Array:
  """Transforms a scalar tensor to a 2 hot representation defined using bins."""
  min_value, max_value = bins[0], bins[-1]
  num_bins = len(bins)
  scalar = jnp.clip(scalar, min_value, max_value)
  upper_index = jnp.argmax(scalar[..., None] <= bins.reshape(1, -1), axis=-1)
  upper_value = bins[upper_index]
  lower_value = bins[upper_index - 1]
  p_lower = (upper_value - scalar) / (upper_value - lower_value)
  p_upper = 1 - p_lower
  lower_one_hot = jax.nn.one_hot(upper_index - 1, num_bins) * p_lower[..., None]
  upper_one_hot = jax.nn.one_hot(upper_index, num_bins) * p_upper[..., None]
  return lower_one_hot + upper_one_hot


def transform_from_2hot_nonlinear(
    probs: Array,
    bins: Array) -> Array:
  """Transforms from a categorical distribution to a scalar."""
  scalar = jnp.sum(probs * jnp.expand_dims(bins, 0), -1)
  return scalar



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `value_learning.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import distributions
from rlax._src import vtrace


class VTraceTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    behavior_policy_logits = np.array(
        [[[8.9, 0.7], [5.0, 1.0], [0.6, 0.1], [-0.9, -0.1]],
         [[0.3, -5.0], [1.0, -8.0], [0.3, 1.7], [4.7, 3.3]]],
        dtype=np.float32)
    target_policy_logits = np.array(
        [[[0.4, 0.5], [9.2, 8.8], [0.7, 4.4], [7.9, 1.4]],
         [[1.0, 0.9], [1.0, -1.0], [-4.3, 8.7], [0.8, 0.3]]],
        dtype=np.float32)
    actions = np.array([[0, 1, 0, 0], [1, 0, 0, 1]], dtype=np.int32)
    self._rho_tm1 = distributions.categorical_importance_sampling_ratios(
        target_policy_logits, behavior_policy_logits, actions)
    self._rewards = np.array(
        [[-1.3, -1.3, 2.3, 42.0],
         [1.3, 5.3, -3.3, -5.0]],
        dtype=np.float32)
    self._discounts = np.array(
        [[0., 0.89, 0.85, 0.99],
         [0.88, 1., 0.83, 0.95]],
        dtype=np.float32)
    self._values = np.array(
        [[2.1, 1.1, -3.1, 0.0],
         [3.1, 0.1, -1.1, 7.4]],
        dtype=np.float32)
    self._bootstrap_value = np.array([8.4, -1.2], dtype=np.float32)
    self._inputs = [
        self._rewards, self._discounts, self._rho_tm1,
        self._values, self._bootstrap_value]

    self._clip_rho_threshold = 1.0
    self._clip_pg_rho_threshold = 5.0
    self._lambda = 1.0

    self._expected_td = np.array(
        [[-1.6155143, -3.4973226, 1.8670533, 5.0316002e1],
         [1.4662437, 3.6116405, -8.3327293e-5, -1.3540000e1]],
        dtype=np.float32)
    self._expected_pg = np.array(
        [[-1.6155143, -3.4973226, 1.8670534, 5.0316002e1],
         [1.4662433, 3.6116405, -8.3369283e-05, -1.3540000e+1]],
        dtype=np.float32)

  @chex.all_variants()
  def test_vtrace_td_error_and_advantage(self):
    """Tests for a full batch."""
    vtrace_td_error_and_advantage = self.variant(jax.vmap(functools.partial(
        vtrace.vtrace_td_error_and_advantage,
        clip_rho_threshold=self._clip_rho_threshold, lambda_=self._lambda)))
    # Get function arguments.
    r_t, discount_t, rho_tm1, v_tm1, bootstrap_value = self._inputs
    v_t = np.concatenate([v_tm1[:, 1:], bootstrap_value[:, None]], axis=1)
    # Compute vtrace output.
    vtrace_output = vtrace_td_error_and_advantage(
        v_tm1, v_t, r_t, discount_t, rho_tm1)
    # Test output.
    np.testing.assert_allclose(
        self._expected_td, vtrace_output.errors, rtol=1e-3)
    np.testing.assert_allclose(
        self._expected_pg, vtrace_output.pg_advantage, rtol=1e-3)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('scalar_lambda',
       np.array([[0., 1., 1., 0., 0., 1., 1., 1.]], dtype=np.float32),
       np.array([1.], dtype=np.float32)),
      ('vector_lambda',
       np.array([[0., 1., 1., 0., 0., 1., 1., 1.]], dtype=np.float32),
       np.array([[1., 1., 1., 1., 1., 1., 1., 1.]], dtype=np.float32)),
      ('vector_lambda_truncation',
       np.array([[0., 1., 1., 1., 0., 1., 1., 1.]], dtype=np.float32),
       np.array([[1., 1., 1., 0., 1., 1., 1., 1.]], dtype=np.float32)),
  )
  def test_vtrace_lambda_multiple_episodes_per_trace(self, discount_t, lambda_):
    """Tests for a full batch."""
    vtrace_ = self.variant(
        jax.vmap(
            functools.partial(
                vtrace.vtrace, clip_rho_threshold=self._clip_rho_threshold)))
    # Get function arguments.
    r_t, rho_tm1, v_tm1 = np.random.random((3, 1, 8))
    bootstrap_value = np.array([10.], dtype=np.float32)
    v_t = np.concatenate([v_tm1[:, 1:], bootstrap_value[:, None]], axis=1)
    # Full trace.
    vtrace_output = vtrace_(v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_)
    # First episode trace.
    vtrace_output_ep1 = vtrace_(v_tm1[:4], v_t[:4], r_t[:4], discount_t[:4],
                                rho_tm1[:4], lambda_[:4])
    # Second episode trace.
    vtrace_output_ep2 = vtrace_(v_tm1[4:], v_t[4:], r_t[4:], discount_t[4:],
                                rho_tm1[4:], lambda_[4:])
    # Test output.
    np.testing.assert_allclose(vtrace_output[:4], vtrace_output_ep1, rtol=1e-3)
    np.testing.assert_allclose(vtrace_output[4:], vtrace_output_ep2, rtol=1e-3)

  @chex.all_variants()
  def test_lambda_q_estimate(self):
    """Tests for a full batch."""
    lambda_ = 0.8
    vtrace_td_error_and_advantage = self.variant(jax.vmap(functools.partial(
        vtrace.vtrace_td_error_and_advantage,
        clip_rho_threshold=self._clip_rho_threshold, lambda_=lambda_)))
    # Get function arguments.
    r_t, discount_t, rho_tm1, v_tm1, bootstrap_value = self._inputs
    v_t = np.concatenate([v_tm1[:, 1:], bootstrap_value[:, None]], axis=1)
    # Compute vtrace output.
    vtrace_output = vtrace_td_error_and_advantage(
        v_tm1, v_t, r_t, discount_t, rho_tm1)
    expected_vs = vtrace_output.errors + v_tm1
    clipped_rho_tm1 = np.minimum(self._clip_rho_threshold, rho_tm1)
    vs_from_q = v_tm1 + clipped_rho_tm1 * (vtrace_output.q_estimate - v_tm1)
    # Test output.
    np.testing.assert_allclose(expected_vs, vs_from_q, rtol=1e-3)

  @chex.all_variants()
  def test_leaky_and_non_leaky_vtrace(self):
    """Tests for a full batch."""
    vtrace_fn = self.variant(jax.vmap(functools.partial(
        vtrace.vtrace, lambda_=self._lambda)))
    leaky_vtrace_fn = self.variant(jax.vmap(functools.partial(
        vtrace.leaky_vtrace, alpha_=1., lambda_=self._lambda)))
    # Get function arguments.
    r_t, discount_t, rho_tm1, v_tm1, bootstrap_value = self._inputs
    v_t = np.concatenate([v_tm1[:, 1:], bootstrap_value[:, None]], axis=1)
    # Compute vtrace and leaky vtrace output.
    vtrace_output = vtrace_fn(v_tm1, v_t, r_t, discount_t, rho_tm1)
    leaky_vtrace_output = leaky_vtrace_fn(v_tm1, v_t, r_t, discount_t, rho_tm1)
    # Test output.
    np.testing.assert_allclose(vtrace_output, leaky_vtrace_output, rtol=1e-3)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for correct handling of interruptions."""

import chex
import dm_env
import jax.numpy as jnp


def fix_step_type_on_interruptions(step_type: chex.Array):
  """Returns step_type with a LAST step before almost every FIRST step.

  If the environment crashes or is interrupted while a trajectory is being
  written, the LAST step can be missing before a FIRST step. We add the LAST
  step before each FIRST step, if the step before the FIRST step is a MID step,
  to signal to the agent that the next observation is not connected to the
  current stream of data. Note that the agent must still then appropriately
  handle both `terminations` (e.g. game over in a game) and `interruptions` (a
  timeout or a reset for system maintenance): the value of the discount on LAST
  step will be > 0 on `interruptions`, while it will be 0 on `terminations`.
  Similar issues arise in hierarchical RL systems as well.

  Args:
    step_type: an array of `dm_env` step types, with shape `[T, B]`.

  Returns:
    Fixed step_type.
  """
  chex.assert_rank(step_type, 2)
  next_step_type = jnp.concatenate([
      step_type[1:],
      jnp.full(
          step_type[:1].shape, int(dm_env.StepType.MID), dtype=step_type.dtype),
  ],
                                   axis=0)
  return jnp.where(
      jnp.logical_and(
          jnp.equal(next_step_type, int(dm_env.StepType.FIRST)),
          jnp.equal(step_type, int(dm_env.StepType.MID)),
      ), int(dm_env.StepType.LAST), step_type)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions implementing custom embeddings."""

import chex
import jax
import jax.numpy as jnp

Array = chex.Array


def embed_oar(features: Array, action: Array, reward: Array,
              num_actions: int) -> Array:
  """Embed each of the (observation, action, reward) inputs & concatenate."""
  chex.assert_rank([features, action, reward], [2, 1, 1])
  action = jax.nn.one_hot(action, num_classes=num_actions)  # [B, A]

  reward = jnp.tanh(reward)
  while reward.ndim < action.ndim:
    reward = jnp.expand_dims(reward, axis=-1)

  embedding = jnp.concatenate([features, action, reward], axis=-1)  # [B, D+A+1]
  return embedding



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `general_value_functions.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import general_value_functions


class PixelControlTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.cell_size = 2
    time, self.batch_size = 3, 2
    height, width, channels = 4, 4, 3

    shape = (time, self.batch_size, height, width, channels)
    hw = np.matmul(
        np.arange(0, 1, 0.25)[:, None],
        np.arange(0, 1, 0.25)[None, :])
    hwc = np.stack([hw, hw + 0.1, hw + 0.2], axis=-1)
    bhwc = np.stack([hwc, hwc + 0.1], axis=0)
    tbhwc = np.stack([bhwc, bhwc + 0.05, bhwc + 0.1], axis=0)
    assert tbhwc.shape == shape
    self.obs = tbhwc

    self.expected = 0.05 * np.ones((2, 2, 2, 2), dtype=np.float32)

  @chex.all_variants()
  def test_pixel_control_rewards(self):
    """Tests for a single element."""
    pixel_control_rewards = self.variant(
        functools.partial(
            general_value_functions.pixel_control_rewards,
            cell_size=self.cell_size))
    # Test pseudo rewards.
    for i in range(self.batch_size):
      rs = pixel_control_rewards(self.obs[:, i])
      np.testing.assert_allclose(self.expected[:, i], rs, rtol=1e-5)

  @chex.all_variants()
  def test_pixel_control_rewards_batch(self):
    """Tests for a batch."""
    pixel_control_rewards = functools.partial(
        general_value_functions.pixel_control_rewards, cell_size=self.cell_size)
    pixel_control_rewards = self.variant(jax.vmap(
        pixel_control_rewards, in_axes=(1,), out_axes=1))
    # Test pseudo rewards.
    rs = pixel_control_rewards(self.obs)
    np.testing.assert_allclose(self.expected, rs, rtol=1e-5)


class FeatureControlTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.potential_discount = .99

    self.features = np.array([
        [[1., 2.], [2., 4.], [3., 6.]],
        [[1., 2.], [0., 0.], [-1., -2.]]])

    self.exp_feature = np.array([
        [[2., 4.], [3., 6.]],
        [[0., 0.], [-1., -2.]]])
    self.exp_abs_change = np.array([
        [[1., 2.], [1., 2.]],
        [[1., 2.], [1., 2.]]])
    self.exp_increase = np.array([
        [[1., 2.], [1., 2.]],
        [[-1., -2.], [-1., -2.]]])
    self.exp_decrease = np.array([
        [[-1., -2.], [-1., -2.]],
        [[1., 2.], [1., 2.]]])

    g = self.potential_discount
    self.exp_potential = np.array([
        [[g*2.-1., g*4.-2.], [g*3.-2., g*6.-4.]],
        [[-1., -2.], [-g, -2.*g]]])

  @chex.all_variants()
  def test_feature_control_rewards_feature_batch(self):
    """Tests for a batch, cumulant_type='feature'."""
    feature_control_rewards = self.variant(jax.vmap(functools.partial(
        general_value_functions.feature_control_rewards,
        cumulant_type='feature')))
    # Test pseudo rewards.
    rs = feature_control_rewards(self.features)
    np.testing.assert_allclose(self.exp_feature, rs, rtol=1e-5)

  @chex.all_variants()
  def test_feature_control_rewards_abs_change_batch(self):
    """Tests for a batch, cumulant_type='absolute_change'."""
    feature_control_rewards = self.variant(jax.vmap(functools.partial(
        general_value_functions.feature_control_rewards,
        cumulant_type='absolute_change')))
    # Test pseudo rewards.
    rs = feature_control_rewards(self.features)
    np.testing.assert_allclose(self.exp_abs_change, rs, rtol=1e-5)

  @chex.all_variants()
  def test_feature_control_rewards_increase_batch(self):
    """Tests for a batch, cumulant_type='increase'."""
    feature_control_rewards = self.variant(jax.vmap(functools.partial(
        general_value_functions.feature_control_rewards,
        cumulant_type='increase')))
    # Test pseudo rewards.
    rs = feature_control_rewards(self.features)
    np.testing.assert_allclose(self.exp_increase, rs, rtol=1e-5)

  @chex.all_variants()
  def test_feature_control_rewards_decrease_batch(self):
    """Tests for a batch, cumulant_type='decrease'."""
    feature_control_rewards = self.variant(jax.vmap(functools.partial(
        general_value_functions.feature_control_rewards,
        cumulant_type='decrease')))
    # Test pseudo rewards.
    rs = feature_control_rewards(self.features)
    np.testing.assert_allclose(self.exp_decrease, rs, rtol=1e-5)

  @chex.all_variants()
  def test_feature_control_rewards_potential_batch(self):
    """Tests for a batch, cumulant_type='potential'."""
    feature_control_rewards = self.variant(jax.vmap(functools.partial(
        general_value_functions.feature_control_rewards,
        cumulant_type='potential',
        discount=self.potential_discount)))
    # Test pseudo rewards.
    rs = feature_control_rewards(self.features)
    np.testing.assert_allclose(self.exp_potential, rs, rtol=1e-5)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Asserts all public symbols are covered in the docs."""

from typing import Any, Mapping

import rlax
from rlax._src import test_utils
from sphinx import application
from sphinx import builders
from sphinx import errors


def rlax_public_symbols():
  names = set()
  for module_name, module in test_utils.find_internal_python_modules(rlax):
    for name in module.__all__:
      names.add(module_name + "." + name)
  return names


class RLaxCoverageCheck(builders.Builder):
  """Builder that checks all public symbols are included."""

  name = "coverage_check"

  def get_outdated_docs(self) -> str:
    return "coverage_check"

  def write(self, *ignored: Any) -> None:
    pass

  def finish(self) -> None:
    documented_objects = frozenset(self.env.domaindata["py"]["objects"])
    undocumented_objects = set(rlax_public_symbols()) - documented_objects
    if undocumented_objects:
      undocumented_objects = tuple(sorted(undocumented_objects))
      raise errors.SphinxError(
          "All public symbols must be included in our documentation, did you "
          "forget to add an entry to `api.rst`?\n"
          f"Undocumented symbols: {undocumented_objects}")


def setup(app: application.Sphinx) -> Mapping[str, Any]:
  app.add_builder(RLaxCoverageCheck)
  return dict(version=rlax.__version__, parallel_read_safe=True)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `policy_gradients.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import distributions
from rlax._src import policy_gradients


class DpgLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.s_t = np.array([[0, 1, 0], [1, 1, 2]], dtype=np.float32)  # [B, T]
    self.w_s = np.ones([3, 2], dtype=np.float32)
    b_s = np.zeros([2], dtype=np.float32)
    # Add batch dimension to satisfy shape assertions.
    self.b_s = jnp.expand_dims(b_s, 0)
    self.w = np.ones([2, 1], dtype=np.float32)
    self.b = np.zeros([1], dtype=np.float32)

    self.expected = np.array([0.5, 0.5], dtype=np.float32)

  @chex.all_variants()
  def test_dpg_loss_batch(self):
    """Tests for a full batch."""
    dpg = self.variant(jax.vmap(functools.partial(
        policy_gradients.dpg_loss, dqda_clipping=1.)))
    # Actor and critic function approximators.
    actor = lambda s_t: jnp.matmul(s_t, self.w_s) + self.b_s
    critic = lambda a_t: jnp.squeeze(jnp.matmul(a_t, self.w) + self.b)
    # Compute loss.
    a_t = actor(self.s_t)
    dqda = jax.vmap(jax.grad(critic))(a_t)
    # Test outputs.
    actual = np.sum(dpg(a_t, dqda), axis=1)
    np.testing.assert_allclose(actual, self.expected, atol=1e-4)


class PolicyGradientLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    logits = np.array(
        [[1., 1., 1.], [2., 0., 0.], [-1., -2., -3.]], dtype=np.float32)
    self.logits = np.stack([logits, logits + 1.])

    weights = np.array([-2., 2., 0], dtype=np.float32)
    self.weights = np.stack([weights, weights - 1.])
    advantages = np.array([0.3, 0.2, 0.1], dtype=np.float32)
    self.advantages = np.stack([advantages, -advantages])
    self.actions = np.array([[0, 1, 2], [0, 0, 0]], dtype=np.int32)

    self.expected = np.array([0.0788835088, 0.327200909], dtype=np.float32)

  @chex.all_variants()
  def test_policy_gradient_loss_batch(self):
    """Tests for a full batch."""
    policy_gradient_loss = self.variant(jax.vmap(
        policy_gradients.policy_gradient_loss))
    # Test outputs.
    actual = policy_gradient_loss(self.logits, self.actions, self.advantages,
                                  self.weights)
    np.testing.assert_allclose(self.expected, actual, atol=1e-4)


class EntropyLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    logits = np.array(
        [[1., 1., 1.], [2., 0., 0.], [-1., -2., -3.]], dtype=np.float32)
    self.logits = np.stack([logits, logits + 1.])
    weights = np.array([-2., 2., 0], dtype=np.float32)
    self.weights = np.stack([weights, weights - 1.])

    self.expected = np.array([0.288693, 1.15422], dtype=np.float32)

  @chex.all_variants()
  def test_entropy_loss_batch(self):
    """Tests for a full batch."""
    entropy_loss = self.variant(jax.vmap(policy_gradients.entropy_loss))
    # Test outputs.
    actual = entropy_loss(self.logits, self.weights)
    np.testing.assert_allclose(self.expected, actual, atol=1e-4)


class QPGLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_values = jnp.array([[0., -1., 1.], [1., -1., 0]])
    self.policy_logits = jnp.array([[1., 1., 1.], [1., 1., 4.]])

    # baseline = \sum_a pi_a * Q_a = 0.
    # -\sum_a pi_a * relu(Q_a - baseline)
    # negative sign as it's a loss term and loss needs to be minimized.
    self.expected_policy_loss = (0.0 + 0.0) / 2

  @parameterized.named_parameters(
      ('JitOnp', jax.jit, lambda t: t),
      ('NoJitOnp', lambda fn: fn, lambda t: t),
      ('JitJnp', jax.jit, jax.device_put),
      ('NoJitJnp', lambda fn: fn, jax.device_put))
  def test_qpg_loss_batch(self, compile_fn, place_fn):
    """Tests for a full batch."""
    # Vmap and optionally compile.
    qpg_loss = compile_fn(policy_gradients.qpg_loss)

    # Optionally convert to device array.
    policy_logits, q_values = jax.tree_map(place_fn,
                                           (self.policy_logits, self.q_values))
    # Test outputs.
    actual = qpg_loss(policy_logits, q_values)
    np.testing.assert_allclose(self.expected_policy_loss, actual, atol=1e-4)


class RMLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.q_values = jnp.array([[0., -1., 1.], [1., -1., 0]])
    self.policy_logits = jnp.array([[1., 1., 1.], [1., 1., 4.]])

    # baseline = \sum_a pi_a * Q_a = 0.
    # -\sum_a pi_a * relu(Q_a - baseline)
    # negative sign as it's a loss term and loss needs to be minimized.
    self.expected_policy_loss = -(.3333 + .0452) / 2

  @parameterized.named_parameters(
      ('JitOnp', jax.jit, lambda t: t),
      ('NoJitOnp', lambda fn: fn, lambda t: t),
      ('JitJnp', jax.jit, jax.device_put),
      ('NoJitJnp', lambda fn: fn, jax.device_put))
  def test_rm_loss_batch(self, compile_fn, place_fn):
    """Tests for a full batch."""
    # Vmap and optionally compile.
    rm_loss = compile_fn(policy_gradients.rm_loss)

    # Optionally convert to device array.
    policy_logits, q_values = jax.tree_map(place_fn,
                                           (self.policy_logits, self.q_values))
    # Test outputs.
    actual = rm_loss(policy_logits, q_values)
    np.testing.assert_allclose(self.expected_policy_loss, actual, atol=1e-4)


class RPGLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_values = jnp.array([[0., -1., 1.], [1., -1., 0]])
    self.policy_logits = jnp.array([[1., 1., 1.], [1., 1., 4.]])

    # baseline = \sum_a pi_a * Q_a = 0.
    # -\sum_a pi_a * relu(Q_a - baseline)
    # negative sign as it's a loss term and loss needs to be minimized.
    self.expected_policy_loss = (1.0 + 1.0) / 2

  @parameterized.named_parameters(
      ('JitOnp', jax.jit, lambda t: t),
      ('NoJitOnp', lambda fn: fn, lambda t: t),
      ('JitJnp', jax.jit, jax.device_put),
      ('NoJitJnp', lambda fn: fn, jax.device_put))
  def test_rpg_loss(self, compile_fn, place_fn):
    """Tests for a full batch."""
    # Vmap and optionally compile.
    rpg_loss = compile_fn(policy_gradients.rpg_loss)

    # Optionally convert to device array.
    policy_logits, q_values = jax.tree_map(place_fn,
                                           (self.policy_logits, self.q_values))
    # Test outputs.
    actual = rpg_loss(policy_logits, q_values)
    np.testing.assert_allclose(self.expected_policy_loss, actual, atol=1e-4)


class ClippedSurrogatePGLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    logits = np.array(
        [[1., 1., 1.], [2., 0., 0.], [-1., -2., -3.]], dtype=np.float32)
    old_logits = np.array(
        [[1., 1., 1.], [2., 0., 0.], [-3., -2., -1.]], dtype=np.float32)
    self.logits = np.stack([logits, logits])
    self.old_logits = np.stack([old_logits, old_logits])

    advantages = np.array([0.3, 0.2, 0.1], dtype=np.float32)
    self.advantages = np.stack([advantages, -advantages])
    self.actions = np.array([[0, 1, 2], [0, 1, 2]], dtype=np.int32)
    self.epsilon = 0.2
    self.expected = np.array([-0.17117467, 0.19333333])

  @chex.all_variants()
  def test_clipped_surrogate_pg_loss_batch(self):
    """Tests for a full batch."""
    get_ratios = jax.vmap(distributions.categorical_importance_sampling_ratios)
    prob_ratios = get_ratios(self.logits, self.old_logits, self.actions)
    batched_fn_variant = self.variant(jax.vmap(functools.partial(
        policy_gradients.clipped_surrogate_pg_loss, epsilon=self.epsilon)))
    actual = batched_fn_variant(prob_ratios, self.advantages)
    np.testing.assert_allclose(actual, self.expected, atol=1e-4)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions implementing different exploration methods.

This file contains a (growing) list of exploration methods used by RL agents.
Currently, we support noise-bsaed exploration methods, such as adding Gaussian
noise or temporally correlated noise drawn from an OU process.

We also support the computation of intrinsic rewards a la Agent57 / NGU style
exploration (see docstring), which is to be used as part of recurrent cell to
process states and a growing memory of previously visited states.
"""
from typing import Optional

import chex
import jax
import jax.numpy as jnp

from rlax._src import episodic_memory

Array = chex.Array
Scalar = chex.Scalar


def add_gaussian_noise(
    key: Array,
    action: Array,
    stddev: float
) -> Array:
  """Returns continuous action with noise drawn from a Gaussian distribution.

  Args:
    key: a key from `jax.random`.
    action: continuous action scalar or vector.
    stddev: standard deviation of noise distribution.

  Returns:
    noisy action, of the same shape as input action.
  """
  chex.assert_type(action, float)

  noise = jax.random.normal(key, shape=action.shape) * stddev
  return action + noise


def add_ornstein_uhlenbeck_noise(
    key: Array,
    action: Array,
    noise_tm1: Array,
    damping: float,
    stddev: float
) -> Array:
  """Returns continuous action with noise from Ornstein-Uhlenbeck process.

  See "On the theory of Brownian Motion" by Uhlenbeck and Ornstein.
  (https://journals.aps.org/pr/abstract/10.1103/PhysRev.36.823).

  Args:
    key: a key from `jax.random`.
    action: continuous action scalar or vector.
    noise_tm1: noise sampled from OU process in previous timestep.
    damping: parameter for controlling autocorrelation of OU process.
    stddev: standard deviation of noise distribution.

  Returns:
    noisy action, of the same shape as input action.
  """
  chex.assert_rank([action, noise_tm1], 1)
  chex.assert_type([action, noise_tm1], float)

  noise_t = (1. - damping) * noise_tm1 + jax.random.normal(
      key, shape=action.shape) * stddev

  return action + noise_t


def add_dirichlet_noise(
    key: Array,
    prior: Array,
    dirichlet_alpha: float,
    dirichlet_fraction: float
) -> Array:
  """Returns discrete actions with noise drawn from a Dirichlet distribution.

  See "Mastering the Game of Go without Human Knowledge" by Silver et. al. 2017
  (https://discovery.ucl.ac.uk/id/eprint/10045895/1/agz_unformatted_nature.pdf),
  "A General Reinforcement Learning Algorithm that Masters Chess, Shogi and
  Go Through Self-Play" by Silver et. al. 2018
  (http://airesearch.com/wp-content/uploads/2016/01/deepmind-mastering-go.pdf),
  and "Mastering Atari, Go, Chess and  Shogi by Planning with a Learned Model"
  by Schrittwieser et. al., 2019 (https://arxiv.org/abs/1911.08265).

  The AlphaZero family of algorithms adds noise sampled from a symmetric
  Dirichlet distribution to the prior policy generated by MCTS. Because the
  agent then samples from this new, noisy prior over actions, this encourages
  better exploration of the root node's children.

  Specifically, this computes:

          noise ~ Dirichlet(alpha)
          noisy_prior = (1 - fraction) * prior + fraction * noise

  Note that alpha is a single float to draw from a symmetric Dirichlet.

  For reference values, AlphaZero uses 0.3, 0.15, 0.03 for Chess, Shogi, and
  Go respectively, and MuZero uses 0.25 for Atari.


  Args:
    key: a key from `jax.random`.
    prior: 2-dim continuous prior policy vector of shapes [B, N], for B batch
      size and N num_actions.
    dirichlet_alpha: concentration parameter to parametrize Dirichlet
      distribution.
    dirichlet_fraction: float from 0 to 1 interpolating between using only the
      prior policy or just the noise.

  Returns:
    noisy action, of the same shape as input action.
  """
  chex.assert_rank(prior, 2)
  chex.assert_type([dirichlet_alpha, dirichlet_fraction], float)

  batch_size, num_actions = prior.shape
  noise = jax.random.dirichlet(
      key=key,
      alpha=jnp.full(shape=(num_actions,), fill_value=dirichlet_alpha),
      shape=(batch_size,))

  noisy_prior = (1 - dirichlet_fraction) * prior + dirichlet_fraction * noise

  return noisy_prior


@chex.dataclass
class IntrinsicRewardState():
  memory: jnp.ndarray
  next_memory_index: Scalar = 0
  distance_sum: Scalar = 0
  distance_count: Scalar = 0


def episodic_memory_intrinsic_rewards(
    embeddings: Array,
    num_neighbors: int,
    reward_scale: float,
    intrinsic_reward_state: Optional[IntrinsicRewardState] = None,
    constant: float = 1e-3,
    epsilon: float = 1e-4,
    cluster_distance: float = 8e-3,
    max_similarity: float = 8.,
    max_memory_size: int = 30_000):
  """Compute intrinsic rewards for exploration via episodic memory.

  This method is adopted from the intrinsic reward computation used in "Never
  Give Up: Learning Directed Exploration Strategies" by Puigdomènech Badia et
  al., (2020) (https://arxiv.org/abs/2003.13350) and "Agent57: Outperforming the
  Atari Human Benchmark" by Puigdomènech Badia et al., (2020)
  (https://arxiv.org/abs/2002.06038).

  From an embedding, we compute the intra-episode intrinsic reward with respect
  to a pre-existing set of embeddings.

  NOTE: For this function to be jittable, static_argnums=[1,] must be passed, as
  the internal jax.lax.top_k(neg_distances, num_neighbors) computation in
  knn_query cannot be jitted with a dynamic num_neighbors that is passed as an
  argument.

  Args:
    embeddings: Array, shaped [M, D] for number of new state embeddings M and
      feature dim D.
    num_neighbors: int for K neighbors used in kNN query
    reward_scale: The β term used in the Agent57 paper to scale the reward.
    intrinsic_reward_state: An IntrinsicRewardState namedtuple, containing
      memory, next_memory_index, distance_sum, and distance_count.
      NOTE- On (only) the first call to episodic_memory_intrinsic_rewards, the
      intrinsic_reward_state is optional, if None is given, an
      IntrinsicRewardState will be initialized with default parameters,
      specifically, the memory will be initialized to an array of jnp.inf of
      shape [max_memory_size x feature dim D], and default values of 0 will be
      provided for next_memory_index, distance_sum, and distance_count.
    constant: float; small constant used for numerical stability used during
      normalizing distances.
    epsilon: float; small constant used for numerical stability when computing
      kernel output.
    cluster_distance: float; the ξ term used in the Agent57 paper to bound the
      distance rate used in the kernel computation.
    max_similarity: float; max limit of similarity; used to zero rewards when
      similarity between memories is too high to be considered 'useful' for an
      agent.
    max_memory_size: int; the maximum number of memories to store. Note that
      performance will be marginally faster if max_memory_size is an exact
      multiple of M (the number of embeddings to add to memory per call to
      episodic_memory_intrinsic_reward).

  Returns:
    Intrinsic reward for each embedding computed by using similarity measure to
    memories and next IntrinsicRewardState.
  """

  # Initialize IntrinsicRewardState if not provided to default values.
  if not intrinsic_reward_state:
    intrinsic_reward_state = IntrinsicRewardState(
        memory=jnp.inf * jnp.ones(shape=(max_memory_size,
                                         embeddings.shape[-1])))
    # Pad the first num_neighbors entries with zeros.
    padding = jnp.zeros((num_neighbors, embeddings.shape[-1]))
    intrinsic_reward_state.memory = (
        intrinsic_reward_state.memory.at[:num_neighbors, :].set(padding))
  else:
    chex.assert_shape(intrinsic_reward_state.memory,
                      (max_memory_size, embeddings.shape[-1]))

  # Compute the KNN from the embeddings using the square distances from
  # the KNN d²(xₖ, x). Results are not guaranteed to be ordered.
  jit_knn_query = jax.jit(episodic_memory.knn_query, static_argnums=[2,])
  knn_query_result = jit_knn_query(intrinsic_reward_state.memory, embeddings,
                                   num_neighbors)

  # Insert embeddings into memory in a ring buffer fashion.
  memory = intrinsic_reward_state.memory
  start_index = intrinsic_reward_state.next_memory_index % memory.shape[0]
  indices = (jnp.arange(embeddings.shape[0]) + start_index) % memory.shape[0]
  memory = jnp.asarray(memory).at[indices].set(embeddings)

  nn_distances_sq = knn_query_result.neighbor_neg_distances

  # Unpack running distance statistics, and update the running mean dₘ²
  distance_sum = intrinsic_reward_state.distance_sum
  distance_sum += jnp.sum(nn_distances_sq)
  distance_counts = intrinsic_reward_state.distance_count
  distance_counts += nn_distances_sq.size

  # We compute the sum of a kernel similarity with the KNN and set to zero
  # the reward when this similarity exceeds a given value (max_similarity)
  # Compute rate = d(xₖ, x)² / dₘ²
  mean_distance = distance_sum / distance_counts
  distance_rate = nn_distances_sq / (mean_distance + constant)

  # The distance rate becomes 0 if already small: r <- max(r-ξ, 0).
  distance_rate = jnp.maximum(distance_rate - cluster_distance,
                              jnp.zeros_like(distance_rate))

  # Compute the Kernel value K(xₖ, x) = ε/(rate + ε).
  kernel_output = epsilon / (distance_rate + epsilon)

  # Compute the similarity for the embedding x:
  # s = √(Σ_{xₖ ∈ Nₖ} K(xₖ, x)) + c
  similarity = (jnp.sqrt(jnp.sum(kernel_output, axis=-1)) + constant)

  # Compute the intrinsic reward:
  # r = 1 / s.
  reward_new = jnp.ones_like(embeddings[..., 0]) / similarity

  # Zero the reward if similarity is greater than max_similarity
  # r <- 0 if s > sₘₐₓ otherwise r.
  max_similarity_reached = similarity > max_similarity
  reward = jnp.where(max_similarity_reached, 0, reward_new)

  # r <- β * r
  reward *= reward_scale

  return reward, IntrinsicRewardState(
      memory=memory,
      next_memory_index=start_index + embeddings.shape[0] % max_memory_size,
      distance_sum=distance_sum,
      distance_count=distance_counts)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for moving_averages.py."""

from absl.testing import absltest
import numpy as np

from rlax._src import moving_averages


class EmaTest(absltest.TestCase):

  def test_moments(self):
    values = [5.0, 7.0]
    decay = 0.9
    d = decay

    init_state, update = moving_averages.create_ema(decay=decay)
    state = init_state(values[0])
    moments, state = update(values[0], state)

    np.testing.assert_allclose(moments.mean, values[0], atol=1e-5)
    np.testing.assert_allclose(moments.variance, 0.0, atol=1e-5)

    moments, state = update(values[1], state)
    np.testing.assert_allclose(
        moments.mean,
        (d * (1 - d) * values[0] + (1 - d) * values[1]) / (1 - d**2), atol=1e-5)
    np.testing.assert_allclose(
        moments.variance,
        (d * (1 - d) * values[0]**2 + (1 - d) * values[1]**2) / (1 - d**2) -
        moments.mean**2, atol=1e-4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A simple double-DQN agent trained to play BSuite's Catch env."""

import collections
import random
from absl import app
from absl import flags
from bsuite.environments import catch
import haiku as hk
from haiku import nets
import jax
import jax.numpy as jnp
import numpy as np
import optax
import rlax
from rlax.examples import experiment

Params = collections.namedtuple("Params", "online target")
ActorState = collections.namedtuple("ActorState", "count")
ActorOutput = collections.namedtuple("ActorOutput", "actions q_values")
LearnerState = collections.namedtuple("LearnerState", "count opt_state")
Data = collections.namedtuple("Data", "obs_tm1 a_tm1 r_t discount_t obs_t")

FLAGS = flags.FLAGS
flags.DEFINE_integer("seed", 42, "Random seed.")
flags.DEFINE_integer("train_episodes", 301, "Number of train episodes.")
flags.DEFINE_integer("batch_size", 32, "Size of the training batch")
flags.DEFINE_float("target_period", 50, "How often to update the target net.")
flags.DEFINE_integer("replay_capacity", 2000, "Capacity of the replay buffer.")
flags.DEFINE_integer("hidden_units", 50, "Number of network hidden units.")
flags.DEFINE_float("epsilon_begin", 1., "Initial epsilon-greedy exploration.")
flags.DEFINE_float("epsilon_end", 0.01, "Final epsilon-greedy exploration.")
flags.DEFINE_integer("epsilon_steps", 1000, "Steps over which to anneal eps.")
flags.DEFINE_float("discount_factor", 0.99, "Q-learning discount factor.")
flags.DEFINE_float("learning_rate", 0.005, "Optimizer learning rate.")
flags.DEFINE_integer("eval_episodes", 100, "Number of evaluation episodes.")
flags.DEFINE_integer("evaluate_every", 50,
                     "Number of episodes between evaluations.")


def build_network(num_actions: int) -> hk.Transformed:
  """Factory for a simple MLP network for approximating Q-values."""

  def q(obs):
    network = hk.Sequential(
        [hk.Flatten(),
         nets.MLP([FLAGS.hidden_units, num_actions])])
    return network(obs)

  return hk.without_apply_rng(hk.transform(q))


class ReplayBuffer(object):
  """A simple Python replay buffer."""

  def __init__(self, capacity):
    self._prev = None
    self._action = None
    self._latest = None
    self.buffer = collections.deque(maxlen=capacity)

  def push(self, env_output, action):
    self._prev = self._latest
    self._action = action
    self._latest = env_output

    if action is not None:
      self.buffer.append(
          (self._prev.observation, self._action, self._latest.reward,
           self._latest.discount, self._latest.observation))

  def sample(self, batch_size):
    obs_tm1, a_tm1, r_t, discount_t, obs_t = zip(
        *random.sample(self.buffer, batch_size))
    return (np.stack(obs_tm1), np.asarray(a_tm1), np.asarray(r_t),
            np.asarray(discount_t) * FLAGS.discount_factor, np.stack(obs_t))

  def is_ready(self, batch_size):
    return batch_size <= len(self.buffer)


class DQN:
  """A simple DQN agent."""

  def __init__(self, observation_spec, action_spec, epsilon_cfg, target_period,
               learning_rate):
    self._observation_spec = observation_spec
    self._action_spec = action_spec
    self._target_period = target_period
    # Neural net and optimiser.
    self._network = build_network(action_spec.num_values)
    self._optimizer = optax.adam(learning_rate)
    self._epsilon_by_frame = optax.polynomial_schedule(**epsilon_cfg)
    # Jitting for speed.
    self.actor_step = jax.jit(self.actor_step)
    self.learner_step = jax.jit(self.learner_step)

  def initial_params(self, key):
    sample_input = self._observation_spec.generate_value()
    sample_input = jnp.expand_dims(sample_input, 0)
    online_params = self._network.init(key, sample_input)
    return Params(online_params, online_params)

  def initial_actor_state(self):
    actor_count = jnp.zeros((), dtype=jnp.float32)
    return ActorState(actor_count)

  def initial_learner_state(self, params):
    learner_count = jnp.zeros((), dtype=jnp.float32)
    opt_state = self._optimizer.init(params.online)
    return LearnerState(learner_count, opt_state)

  def actor_step(self, params, env_output, actor_state, key, evaluation):
    obs = jnp.expand_dims(env_output.observation, 0)  # add dummy batch
    q = self._network.apply(params.online, obs)[0]  # remove dummy batch
    epsilon = self._epsilon_by_frame(actor_state.count)
    train_a = rlax.epsilon_greedy(epsilon).sample(key, q)
    eval_a = rlax.greedy().sample(key, q)
    a = jax.lax.select(evaluation, eval_a, train_a)
    return ActorOutput(actions=a, q_values=q), ActorState(actor_state.count + 1)

  def learner_step(self, params, data, learner_state, unused_key):
    target_params = optax.periodic_update(params.online, params.target,
                                          learner_state.count,
                                          self._target_period)
    dloss_dtheta = jax.grad(self._loss)(params.online, target_params, *data)
    updates, opt_state = self._optimizer.update(dloss_dtheta,
                                                learner_state.opt_state)
    online_params = optax.apply_updates(params.online, updates)
    return (Params(online_params, target_params),
            LearnerState(learner_state.count + 1, opt_state))

  def _loss(self, online_params, target_params, obs_tm1, a_tm1, r_t, discount_t,
            obs_t):
    q_tm1 = self._network.apply(online_params, obs_tm1)
    q_t_val = self._network.apply(target_params, obs_t)
    q_t_select = self._network.apply(online_params, obs_t)
    batched_loss = jax.vmap(rlax.double_q_learning)
    td_error = batched_loss(q_tm1, a_tm1, r_t, discount_t, q_t_val, q_t_select)
    return jnp.mean(rlax.l2_loss(td_error))


def main(unused_arg):
  env = catch.Catch(seed=FLAGS.seed)
  epsilon_cfg = dict(
      init_value=FLAGS.epsilon_begin,
      end_value=FLAGS.epsilon_end,
      transition_steps=FLAGS.epsilon_steps,
      power=1.)
  agent = DQN(
      observation_spec=env.observation_spec(),
      action_spec=env.action_spec(),
      epsilon_cfg=epsilon_cfg,
      target_period=FLAGS.target_period,
      learning_rate=FLAGS.learning_rate,
  )

  accumulator = ReplayBuffer(FLAGS.replay_capacity)
  experiment.run_loop(
      agent=agent,
      environment=env,
      accumulator=accumulator,
      seed=FLAGS.seed,
      batch_size=FLAGS.batch_size,
      train_episodes=FLAGS.train_episodes,
      evaluate_every=FLAGS.evaluate_every,
      eval_episodes=FLAGS.eval_episodes,
  )


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for perturbations.py."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import nested_updates


class NestedUpdatesTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    old = jnp.zeros((3,), dtype=jnp.float32)
    new = jnp.ones((3,), dtype=jnp.float32)
    self._old_struct = ((old, old), old)
    self._new_struct = ((new, new), new)

  @chex.all_variants()
  def test_conditional_update_is_time(self):
    """Check periodic update enabled."""
    conditional_update = self.variant(nested_updates.conditional_update)

    is_time = jnp.array(True)
    output = conditional_update(self._new_struct, self._old_struct, is_time)
    for o, exp in zip(
        jax.tree_leaves(output), jax.tree_leaves(self._new_struct)):
      np.testing.assert_allclose(o, exp)

  @chex.all_variants()
  def test_conditional_update_is_not_time(self):
    """Check periodic update disables."""
    conditional_update = self.variant(nested_updates.conditional_update)

    is_not_time = jnp.array(False)
    output = conditional_update(self._new_struct, self._old_struct, is_not_time)
    for o, exp in zip(
        jax.tree_leaves(output), jax.tree_leaves(self._old_struct)):
      np.testing.assert_allclose(o, exp)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Common utilities for RLax functions."""

from typing import Optional, Sequence, Union
import chex
import jax
import jax.numpy as jnp
import numpy as np
Array = chex.Array
Numeric = chex.Numeric


def batched_index(
    values: Array, indices: Array, keepdims: bool = False
) -> Array:
  """Index into the last dimension of a tensor, preserving all others dims.

  Args:
    values: a tensor of shape [..., D],
    indices: indices of shape [...].
    keepdims: whether to keep the final dimension.

  Returns:
    a tensor of shape [...] or [..., 1].
  """
  indexed = jnp.take_along_axis(values, indices[..., None], axis=-1)
  if not keepdims:
    indexed = jnp.squeeze(indexed, axis=-1)
  return indexed


def one_hot(indices, num_classes, dtype=jnp.float32):
  """Returns a one-hot version of indices.

  Args:
    indices: A tensor of indices.
    num_classes: Number of classes in the one-hot dimension.
    dtype: The dtype.

  Returns:
    The one-hot tensor. If indices' shape is [A, B, ...], shape is
    [A, B, ..., num_classes].
  """
  labels = jnp.arange(num_classes)
  for _ in range(indices.ndim):
    labels = jnp.expand_dims(labels, axis=0)
  return jnp.array(
      indices[..., jnp.newaxis] == labels, dtype=dtype)


def lhs_broadcast(source, target):
  """Ensures that source is compatible with target for broadcasting."""
  same_shape = np.array(source.shape) == np.array(target.shape[:source.ndim])
  ones = np.array(source.shape) == np.ones((source.ndim,))
  if np.all(same_shape + ones):
    broadcast_shape = source.shape + (1,) * (target.ndim - source.ndim)
    return jnp.reshape(source, broadcast_shape)
  raise ValueError(
      f"source shape {source.shape} is not compatible with "
      f"target shape {target.shape}")


def replace_masked(data: chex.Array, replacement: chex.Array, mask: chex.Array):
  """Replace slices of an array as indicated by a mask.

  Args:
    data: an array whose some elements we want to replace.
    replacement: an array with the same shape as `data`, containing
      the data to insert according to `mask`. If `None` is passed,
      then the masked elements in `data` will be replaced with zeros.
    mask: a mask of 0/1s, whose shape is a prefix of `data` and `replacements`.
      When the mask is 1, values of data are replaced.

  Returns:
    the updated tensor.
  """
  if replacement is None:
    replacement = jnp.zeros_like(data)
  return jnp.where(lhs_broadcast(mask, data), replacement, data)


class AllSum:
  """Helper for summing over elements in an array and over devices."""

  def __init__(self, axis_name: Optional[str] = None):
    """Sums locally and then over devices with the axis name provided."""
    self._axis_name = axis_name

  def __call__(
      self, x: Array, axis: Optional[Union[int, Sequence[int]]] = None
  ) -> Numeric:
    s = jnp.sum(x, axis=axis)
    if self._axis_name:
      s = jax.lax.psum(s, axis_name=self._axis_name)
    return s



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for episodic_memory.py."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import episodic_memory


class KNNQueryTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.data = np.array([[0., 0.], [7.5, 1.], [40., 40.]])
    self.query_points = np.array([[2.0, 1.3], [7.5, 0.0]])

  @chex.all_variants()
  def test_small_k_query(self):
    num_neighbors = 2
    expected_neighbors = np.array([[[0., 0.], [7.5, 1.]],
                                   [[7.5, 1.], [0., 0.]]])
    expected_neg_distances = np.array([[-5.69, -30.34], [-1., -56.25]])
    expected_neighbor_indices = np.array([[0., 1.], [1., 0.]])

    @self.variant
    def query_variant(data, points):
      return episodic_memory.knn_query(data, points, num_neighbors)
    actual = query_variant(self.data, self.query_points)

    np.testing.assert_allclose(actual.neighbors,
                               expected_neighbors,
                               atol=1e-6)
    np.testing.assert_allclose(actual.neighbor_indices,
                               expected_neighbor_indices,
                               atol=1e-6)
    np.testing.assert_allclose(actual.neighbor_neg_distances,
                               expected_neg_distances,
                               atol=1e-6)

  @chex.all_variants()
  @parameterized.named_parameters(('3neighbors', 3),
                                  ('5neighbors', 5))
  def test_big_k_query(self, num_neighbors):
    expected_neighbors = np.array([[[0., 0.], [7.5, 1.], [40., 40.]],
                                   [[7.5, 1.], [0., 0.], [40., 40.]]])
    expected_neg_distances = np.array([[-5.69, -30.34, -2941.69],
                                       [-1., -56.25, -2656.25]])
    expected_neighbor_indices = np.array([[0, 1, 2], [1, 0, 2],])

    @self.variant
    def query_variant(data, points):
      return episodic_memory.knn_query(data, points, num_neighbors)
    actual = query_variant(self.data, self.query_points)

    np.testing.assert_allclose(actual.neighbors,
                               expected_neighbors,
                               atol=1e-6)
    np.testing.assert_allclose(actual.neighbor_indices,
                               expected_neighbor_indices,
                               atol=1e-6)
    np.testing.assert_allclose(actual.neighbor_neg_distances,
                               expected_neg_distances,
                               atol=1e-6)

  def test_empty(self):
    data = np.array([])
    self.query_points = np.array([])
    with self.assertRaises(AssertionError):
      episodic_memory.knn_query(data, self.query_points, num_neighbors=2)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""RLax: building blocks for RL, in JAX."""

from rlax._src.base import AllSum
from rlax._src.base import batched_index
from rlax._src.base import lhs_broadcast
from rlax._src.base import one_hot
from rlax._src.base import replace_masked
from rlax._src.clipping import clip_gradient
from rlax._src.clipping import huber_loss
from rlax._src.distributions import categorical_cross_entropy
from rlax._src.distributions import categorical_importance_sampling_ratios
from rlax._src.distributions import categorical_kl_divergence
from rlax._src.distributions import categorical_sample
from rlax._src.distributions import clipped_entropy_softmax
from rlax._src.distributions import epsilon_greedy
from rlax._src.distributions import gaussian_diagonal
from rlax._src.distributions import greedy
from rlax._src.distributions import multivariate_normal_kl_divergence
from rlax._src.distributions import softmax
from rlax._src.distributions import squashed_gaussian
from rlax._src.embedding import embed_oar
from rlax._src.episodic_memory import knn_query
from rlax._src.exploration import add_dirichlet_noise
from rlax._src.exploration import add_gaussian_noise
from rlax._src.exploration import add_ornstein_uhlenbeck_noise
from rlax._src.exploration import episodic_memory_intrinsic_rewards
from rlax._src.general_value_functions import feature_control_rewards
from rlax._src.general_value_functions import pixel_control_rewards
from rlax._src.interruptions import fix_step_type_on_interruptions
from rlax._src.losses import expectile_loss
from rlax._src.losses import l2_loss
from rlax._src.losses import likelihood
from rlax._src.losses import log_loss
from rlax._src.losses import pixel_control_loss
from rlax._src.model_learning import extract_subsequences
from rlax._src.model_learning import sample_start_indices
from rlax._src.moving_averages import create_ema
from rlax._src.moving_averages import EmaMoments
from rlax._src.moving_averages import EmaState
from rlax._src.mpo_ops import compute_parametric_kl_penalty_and_dual_loss
from rlax._src.mpo_ops import LagrangePenalty
from rlax._src.mpo_ops import mpo_compute_weights_and_temperature_loss
from rlax._src.mpo_ops import mpo_loss
from rlax._src.mpo_ops import vmpo_compute_weights_and_temperature_loss
from rlax._src.mpo_ops import vmpo_loss
from rlax._src.multistep import discounted_returns
from rlax._src.multistep import general_off_policy_returns_from_action_values
from rlax._src.multistep import general_off_policy_returns_from_q_and_v
from rlax._src.multistep import lambda_returns
from rlax._src.multistep import n_step_bootstrapped_returns
from rlax._src.multistep import truncated_generalized_advantage_estimation
from rlax._src.nested_updates import conditional_update
from rlax._src.nested_updates import periodic_update
from rlax._src.nonlinear_bellman import compose_tx
from rlax._src.nonlinear_bellman import DISCOUNT_TRANSFORM_PAIR
from rlax._src.nonlinear_bellman import HYPERBOLIC_SIN_PAIR
from rlax._src.nonlinear_bellman import IDENTITY_PAIR
from rlax._src.nonlinear_bellman import muzero_pair
from rlax._src.nonlinear_bellman import SIGNED_HYPERBOLIC_PAIR
from rlax._src.nonlinear_bellman import SIGNED_LOGP1_PAIR
from rlax._src.nonlinear_bellman import transformed_general_off_policy_returns_from_action_values
from rlax._src.nonlinear_bellman import transformed_lambda_returns
from rlax._src.nonlinear_bellman import transformed_n_step_q_learning
from rlax._src.nonlinear_bellman import transformed_n_step_returns
from rlax._src.nonlinear_bellman import transformed_q_lambda
from rlax._src.nonlinear_bellman import transformed_retrace
from rlax._src.nonlinear_bellman import twohot_pair
from rlax._src.nonlinear_bellman import TxPair
from rlax._src.nonlinear_bellman import unbiased_transform_pair
from rlax._src.policy_gradients import clipped_surrogate_pg_loss
from rlax._src.policy_gradients import dpg_loss
from rlax._src.policy_gradients import entropy_loss
from rlax._src.policy_gradients import policy_gradient_loss
from rlax._src.policy_gradients import qpg_loss
from rlax._src.policy_gradients import rm_loss
from rlax._src.policy_gradients import rpg_loss
from rlax._src.policy_targets import constant_policy_targets
from rlax._src.policy_targets import PolicyTarget
from rlax._src.policy_targets import sampled_policy_distillation_loss
from rlax._src.policy_targets import zero_policy_targets
from rlax._src.pop_art import art
from rlax._src.pop_art import normalize
from rlax._src.pop_art import pop
from rlax._src.pop_art import popart
from rlax._src.pop_art import PopArtState
from rlax._src.pop_art import unnormalize
from rlax._src.pop_art import unnormalize_linear
from rlax._src.transforms import identity
from rlax._src.transforms import logit
from rlax._src.transforms import power
from rlax._src.transforms import sigmoid
from rlax._src.transforms import signed_expm1
from rlax._src.transforms import signed_hyperbolic
from rlax._src.transforms import signed_logp1
from rlax._src.transforms import signed_parabolic
from rlax._src.transforms import transform_from_2hot
from rlax._src.transforms import transform_to_2hot
from rlax._src.tree_util import transpose_first_axis_to_last
from rlax._src.tree_util import transpose_last_axis_to_first
from rlax._src.tree_util import tree_fn
from rlax._src.tree_util import tree_map_zipped
from rlax._src.tree_util import tree_replace_masked
from rlax._src.tree_util import tree_select
from rlax._src.tree_util import tree_split_key
from rlax._src.tree_util import tree_split_leaves
from rlax._src.value_learning import categorical_double_q_learning
from rlax._src.value_learning import categorical_l2_project
from rlax._src.value_learning import categorical_q_learning
from rlax._src.value_learning import categorical_td_learning
from rlax._src.value_learning import double_q_learning
from rlax._src.value_learning import expected_sarsa
from rlax._src.value_learning import persistent_q_learning
from rlax._src.value_learning import q_lambda
from rlax._src.value_learning import q_learning
from rlax._src.value_learning import quantile_expected_sarsa
from rlax._src.value_learning import quantile_q_learning
from rlax._src.value_learning import quantile_regression_loss
from rlax._src.value_learning import qv_learning
from rlax._src.value_learning import qv_max
from rlax._src.value_learning import retrace
from rlax._src.value_learning import retrace_continuous
from rlax._src.value_learning import sarsa
from rlax._src.value_learning import sarsa_lambda
from rlax._src.value_learning import td_lambda
from rlax._src.value_learning import td_learning
from rlax._src.vtrace import leaky_vtrace
from rlax._src.vtrace import leaky_vtrace_td_error_and_advantage
from rlax._src.vtrace import vtrace
from rlax._src.vtrace import vtrace_td_error_and_advantage

__version__ = "0.1.5"

__all__ = (
    "add_gaussian_noise",
    "add_ornstein_uhlenbeck_noise",
    "add_dirichlet_noise",
    "AllSum",
    "batched_index",
    "categorical_cross_entropy",
    "categorical_double_q_learning",
    "categorical_importance_sampling_ratios",
    "categorical_kl_divergence",
    "categorical_l2_project",
    "categorical_q_learning",
    "categorical_td_learning",
    "clip_gradient",
    "clipped_surrogate_pg_loss",
    "compose_tx",
    "conditional_update",
    "constant_policy_targets",
    "create_ema",
    "discounted_returns",
    "DISCOUNT_TRANSFORM_PAIR",
    "double_q_learning",
    "dpg_loss",
    "EmaMoments",
    "EmaState",
    "entropy_loss",
    "episodic_memory_intrinsic_rewards",
    "epsilon_greedy",
    "expected_sarsa",
    "expectile_loss",
    "extract_subsequences",
    "feature_control_rewards",
    "fix_step_type_on_interruptions",
    "gaussian_diagonal",
    "HYPERBOLIC_SIN_PAIR",
    "squashed_gaussian",
    "clipped_entropy_softmax",
    "art",
    "compute_parametric_kl_penalty_and_dual_loss",
    "general_off_policy_returns_from_action_values",
    "general_off_policy_returns_from_q_and_v",
    "greedy",
    "huber_loss",
    "identity",
    "IDENTITY_PAIR",
    "knn_query",
    "l2_loss",
    "LagrangePenalty",
    "lambda_returns",
    "leaky_vtrace",
    "leaky_vtrace_td_error_and_advantage",
    "lhs_broadcast",
    "likelihood",
    "logit",
    "log_loss",
    "mpo_compute_weights_and_temperature_loss",
    "mpo_loss",
    "multivariate_normal_kl_divergence",
    "muzero_pair",
    "normalize",
    "n_step_bootstrapped_returns",
    "one_hot",
    "periodic_update",
    "persistent_q_learning",
    "pixel_control_rewards",
    "PolicyTarget",
    "policy_gradient_loss",
    "pop",
    "popart",
    "PopArtState",
    "power",
    "qpg_loss",
    "quantile_expected_sarsa",
    "quantile_q_learning",
    "quantile_regression_loss",
    "qv_learning",
    "qv_max",
    "q_lambda",
    "q_learning",
    "replace_masked",
    "retrace",
    "retrace_continuous",
    "rm_loss",
    "rpg_loss",
    "sample_start_indices",
    "sampled_policy_distillation_loss",
    "sarsa",
    "sarsa_lambda",
    "sigmoid",
    "signed_expm1",
    "signed_hyperbolic",
    "SIGNED_HYPERBOLIC_PAIR",
    "signed_logp1",
    "SIGNED_LOGP1_PAIR",
    "signed_parabolic",
    "softmax",
    "td_lambda",
    "td_learning",
    "transformed_general_off_policy_returns_from_action_values",
    "transformed_lambda_returns",
    "transformed_n_step_q_learning",
    "transformed_n_step_returns",
    "transformed_q_lambda",
    "transformed_retrace",
    "transform_from_2hot",
    "transform_to_2hot",
    "transpose_last_axis_to_first",
    "transpose_first_axis_to_last",
    "tree_fn",
    "tree_map_zipped",
    "tree_replace_masked",
    "tree_select",
    "tree_split_key",
    "tree_split_leaves",
    "truncated_generalized_advantage_estimation",
    "twohot_pair",
    "TxPair",
    "unbiased_transform_pair",
    "unnormalize",
    "unnormalize_linear",
    "vmpo_compute_weights_and_temperature_loss",
    "vmpo_loss",
    "vtrace",
    "vtrace_td_error_and_advantage",
    "zero_policy_targets",
)

#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the RLax public API.    /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions to support model learning."""

from typing import Optional

import chex
import jax
import jax.numpy as jnp


def extract_subsequences(
    trajectories: chex.Array,
    start_indices: chex.Array,
    subsequence_len: int = 1,
    max_valid_start_idx: Optional[int] = None,
) -> chex.Array:
  """Extract (potentially overlapping) subsequences from batch of trajectories.

  WARNING: If `max_valid_start_idx` is not provided, or incorrectly set,
  the function cannot check the validity of the chosen `start_idx` and no error
  will be raised if indexing outside the data boundaries.

  Args:
    trajectories: A batch of trajectories, shape `[T, B, ...]`.
    start_indices: Time indices of start points, shape `[B, num_start_indices]`.
    subsequence_len: The length of subsequences extracted from `trajectories`.
    max_valid_start_idx: the maximum valid start index, therefore the
      `start_indices` should be from {0, ..., max_valid_start_idx}.

  Returns:
    A batch of subsequences, with
    `trajectories[start_indices[i, j]:start_indices[i, j] + n]` for each start
    index. Output shape is: `[subsequence_len, B, num_start_indices, ...]`.
  """
  if max_valid_start_idx is not None:
    min_len = max_valid_start_idx + subsequence_len
    traj_len = trajectories.shape[0]
    if traj_len < min_len:
      raise AssertionError(
          f'Expected len >= {min_len}, but trajectories length is: {traj_len}.')

  batch_size = start_indices.shape[0]
  batch_range = jnp.arange(batch_size)
  num_subs = start_indices.shape[1]
  idx_arr = jnp.arange(subsequence_len)[:, None, None] * jnp.ones(
      (subsequence_len, batch_size, num_subs), dtype=jnp.int32) + start_indices
  return trajectories[idx_arr, batch_range[None, :, None], ...]


def sample_start_indices(
    rng_key: chex.PRNGKey,
    batch_size: int,
    num_start_indices: int,
    max_valid_start_idx: int
) -> chex.Array:
  """Sampling `batch_size x num_start_indices` starting indices.

  Args:
    rng_key: a pseudo random number generator's key.
    batch_size: the size of the batch of trajectories to index in.
    num_start_indices: how many starting points per trajectory in the batch.
    max_valid_start_idx: maximum valid time index for all starting points.

  Returns:
    an array of starting points with shape `[B, num_start_indices]`
  """

  @jax.vmap
  def _vchoose(key, entries):
    return jax.random.choice(
        key, entries, shape=(num_start_indices,), replace=False)

  rollout_window = jnp.arange(max_valid_start_idx + 1)
  return _vchoose(
      jax.random.split(rng_key, batch_size),
      jnp.tile(rollout_window, (batch_size, 1)))



# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Maximum A Posteriori Policy Optimization (MPO/V-MPO) ops.

  Maximum a Posteriori Policy Optimisation.
  https://openreview.net/forum?id=S1ANxQW0b

  Relative Entropy Regularized Policy Iteration.
  https://arxiv.org/abs/1812.02256

  V-MPO: On-Policy Maximum a Posteriori Policy Optimization
  for Discrete and Continuous Control.
  https://openreview.net/forum?id=SylOlp4FvH

Since these functions are calculated per-example (with some aggregation over
all examples), they work with many input shapes as long
as input shapes are consistent across inputs. We use E* to denote the shape of
the examples. For example, E* could be [T, B], [B, T], [T], etc as long as E* is
consistent across all function inputs, and function output shape will also
depend on E*.
"""
import functools
from typing import Callable, NamedTuple, Optional, Sequence, Tuple

from absl import logging
import chex
import jax
import jax.numpy as jnp
from rlax._src import base

# This op is not in the list of officially supported ops because the jax team
# have not fully tested it, but it works nonetheless so we add and use it.
# TODO(b/160450576): Remove when this op is officially supported.
jax.interpreters.pxla.multi_host_supported_collectives.add(jax.lax.pmax_p)

Array = chex.Array
Numeric = chex.Numeric
Scalar = chex.Scalar


class LagrangePenalty(NamedTuple):
  # Dual variable responsible for modulating the penalty for this constraint.
  alpha: Array
  # Bound for this constraint.
  epsilon: Scalar
  # Whether to constrain each dimension separately with its own bound epsilon.
  per_dimension: bool = False


class MpoOutputs(NamedTuple):
  """Additional outputs for mpo loss functions."""
  # In VMPO temperature loss is computed across all data so should be scalar, in
  # MPO this is per example with shape E*
  temperature_loss: Numeric
  # These losses are per example with shape E*.
  policy_loss: Array
  kl_loss: Array
  alpha_loss: Array
  # Other outputs.
  normalized_weights: Array
  num_samples: Numeric

_EPSILON = 1e-10  # For numerical stability.
_INFINITY = 1e6


def mpo_loss(
    sample_log_probs: Array,
    sample_q_values: Array,
    temperature_constraint: LagrangePenalty,
    kl_constraints: Sequence[Tuple[Array, LagrangePenalty]],
    projection_operator: Callable[[Numeric], Numeric] = functools.partial(
        jnp.clip, a_min=_EPSILON),
    policy_loss_weight: float = 1.0,
    temperature_loss_weight: float = 1.0,
    kl_loss_weight: float = 1.0,
    alpha_loss_weight: float = 1.0,
    sample_axis: int = 0,
    use_stop_gradient: bool = True,
) -> Tuple[Array, MpoOutputs]:
  """Implements the MPO loss with a KL bound.

  This loss implements the MPO algorithm for policies with a bound for the KL
  between the current and target policy.

  Note: This is a per-example loss which works on any shape inputs as long as
  they are consistent. We denote this shape E* for ease of reference. Args
  sample_log_probs and sample_q_values are shape E + an extra sample axis that
  contains the sampled actions' log probs and q values respectively. For
  example, if sample_axis = 0, the shapes expected will be [S, E*]. Or if
  E* = [T, B] and sample_axis = 1, the shapes expected will be [T, S, B].

  Args:
    sample_log_probs: An array of shape E* + a sample axis inserted at
      sample_axis containing the log probabilities of the sampled actions under
      the current policy.
    sample_q_values: An array of shape E* + a sample axis inserted at
      sample_axis containing the q function values evaluated on the sampled
      actions.
    temperature_constraint: Lagrange constraint for the E-step temperature
      optimization.
    kl_constraints: KL and variables for applying Lagrangian penalties to bound
      them in the M-step, KLs are  [E*, A?]. Here A is the action dimension
      in the case of per-dimension KL constraints.
    projection_operator: Function to project dual variables (temperature and kl
      constraint alphas) into the positive range.
    policy_loss_weight: Weight for the policy loss.
    temperature_loss_weight: Weight for the temperature loss.
    kl_loss_weight: Weight for the KL loss.
    alpha_loss_weight: Weight for the alpha loss.
    sample_axis: Axis in sample_log_probs and sample_q_values that contains the
      sampled actions' log probs and q values respectively. For example, if
      sample_axis = 0, the shapes expected will be [S, E*]. Or if E* = [T, B]
      and sample_axis = 1, the shapes expected will be [T, S, B].
    use_stop_gradient: bool indicating whether or not to apply stop gradient.

  Returns:
    Per example `loss` with shape E*, and additional data including
    the components of this loss and the normalized weights in the
    AdditionalOutputs.
  """
  chex.assert_equal_shape([sample_log_probs, sample_q_values])
  chex.assert_rank(temperature_constraint.epsilon, 0)
  chex.assert_type([
      sample_log_probs, sample_q_values, temperature_constraint.alpha,
      temperature_constraint.epsilon], float)

  for kl, penalty in kl_constraints:
    chex.assert_rank(penalty.epsilon, 0)
    chex.assert_type([kl, penalty.alpha, penalty.epsilon], float)
    if penalty.per_dimension:
      chex.assert_rank(kl, sample_q_values.ndim)
    else:
      chex.assert_rank(kl, sample_q_values.ndim - 1)

  if sample_axis < 0:
    sample_axis += sample_q_values.ndim
  if not 0 <= sample_axis < sample_q_values.ndim:
    raise ValueError(
        f"`sample_axis` {sample_axis} not in array rank {sample_q_values.ndim}")

  # E-Step. Compute temperature loss, weights, and temperature.
  temperature_loss, norm_weights, num_samples = (
      mpo_compute_weights_and_temperature_loss(
          sample_q_values, temperature_constraint, projection_operator,
          sample_axis=sample_axis))

  norm_weights = jax.lax.select(
      use_stop_gradient, jax.lax.stop_gradient(norm_weights), norm_weights)

  # M-Step. Supervised learning on reweighted probabilities using the weights
  # from the E-Step under an additional KL constraint.
  policy_loss = -jnp.sum(norm_weights * sample_log_probs, axis=sample_axis)
  kl_loss, alpha_loss = compute_parametric_kl_penalty_and_dual_loss(
      kl_constraints, projection_operator, use_stop_gradient)

  chex.assert_equal_shape([policy_loss, kl_loss, alpha_loss])

  # Combine all loss components. The final loss is of shape E*.
  loss = (policy_loss_weight * policy_loss +
          temperature_loss_weight * temperature_loss +
          kl_loss_weight * kl_loss +
          alpha_loss_weight * alpha_loss)
  return loss, MpoOutputs(
      temperature_loss=temperature_loss, policy_loss=policy_loss,
      kl_loss=kl_loss, alpha_loss=alpha_loss, normalized_weights=norm_weights,
      num_samples=num_samples)


def mpo_compute_weights_and_temperature_loss(
    sample_q_values: Array,
    temperature_constraint: LagrangePenalty,
    projection_operator: Callable[[Numeric], Numeric],
    sample_axis: int = 0,
) -> Tuple[Array, Array, Scalar]:
  """Computes the weights and temperature loss for MPO.

  The E-Step computes a non-parameteric sample-based approximation of the
  current policy by reweighting the state-action value function.

  Here, we compute this nonparametric policy and optimize the temperature
  parameter used in the reweighting.

  Args:
    sample_q_values: An array of shape E* + a sample axis inserted at
      sample_axis containing the q function values evaluated on the sampled
      actions.
    temperature_constraint: Lagrange constraint for the E-step temperature
      optimization.
    projection_operator: Function to project temperature into the positive
      range.
    sample_axis: Axis in sample_q_values containing sampled actions.

  Returns:
    The temperature loss, normalized weights and number of actions samples per
    state.
  """
  chex.assert_rank(temperature_constraint.epsilon, 0)
  chex.assert_type([sample_q_values, temperature_constraint.alpha,
                    temperature_constraint.epsilon], float)

  if sample_axis < 0:
    sample_axis += sample_q_values.ndim
  if not 0 <= sample_axis < sample_q_values.ndim:
    raise ValueError(
        f"`sample_axis` {sample_axis} not in array rank {sample_q_values.ndim}")

  n_action_samples = sample_q_values.shape[sample_axis]

  # Clip the temperature value (temperature must be positive).
  temperature = projection_operator(temperature_constraint.alpha)
  epsilon = temperature_constraint.epsilon

  # Scale the Q-values.
  scaled_sample_q_values = sample_q_values / temperature

  # Temperature optimization.
  q_logsumexp = jax.scipy.special.logsumexp(
      scaled_sample_q_values, axis=sample_axis, keepdims=True)

  # The temperature loss encourages the current and previous policy to stay
  # close. This loss optimizes the convex dual of an upper bound on the average
  # KL (epsilon) between the current and previous state-action values.
  temperature_loss = (
      temperature * epsilon +
      (temperature * (jnp.squeeze(q_logsumexp, axis=sample_axis)
                      - jnp.log(n_action_samples))))

  # The weights corresponds to a softmax over state-action values.
  weights = jnp.exp(scaled_sample_q_values - q_logsumexp)

  # Normalize the weights before the M-Step
  norm_weights = weights / jnp.sum(weights, axis=sample_axis, keepdims=True)

  return temperature_loss, norm_weights, n_action_samples


def compute_parametric_kl_penalty_and_dual_loss(
    kl_constraints: Sequence[Tuple[Array, LagrangePenalty]],
    projection_operator: Callable[[Numeric], Numeric],
    use_stop_gradient: bool = True,
) -> Tuple[Array, Array]:
  """Optimize hard KL constraints between the current and previous policies."""
  for kl, penalty in kl_constraints:
    chex.assert_rank(penalty.epsilon, 0)
    chex.assert_type([kl, penalty.alpha, penalty.epsilon], float)

  kl_losses, alpha_losses = [], []
  for kl, penalty in kl_constraints:
    kl_loss, alpha_loss, _ = kl_constraint_loss(
        kl, penalty, projection_operator, use_stop_gradient)
    kl_losses.append(kl_loss)
    alpha_losses.append(alpha_loss)
  kl_loss, alpha_loss = sum(kl_losses), sum(alpha_losses)
  return kl_loss, alpha_loss


def vmpo_loss(
    sample_log_probs: Array,
    advantages: Array,
    temperature_constraint: LagrangePenalty,
    kl_constraints: Sequence[Tuple[Array, LagrangePenalty]],
    projection_operator: Callable[[Numeric], Numeric] = functools.partial(
        jnp.clip, a_min=_EPSILON),
    restarting_weights: Optional[Array] = None,
    importance_weights: Optional[Array] = None,
    top_k_fraction: float = 0.5,
    policy_loss_weight: float = 1.0,
    temperature_loss_weight: float = 1.0,
    kl_loss_weight: float = 1.0,
    alpha_loss_weight: float = 1.0,
    axis_name: Optional[str] = None,
    use_stop_gradient: bool = True,
) -> Tuple[Array, MpoOutputs]:
  """Calculates the V-MPO policy improvement loss.

  Note: This is a per-example loss which works on any shape inputs as long as
  they are consistent. We denote the shape of the examples E* for ease of
  reference.

  Args:
    sample_log_probs: Log probabilities of actions for each example. Shape E*.
    advantages: Advantages for the E-step. Shape E*.
    temperature_constraint: Lagrange constraint for the E-step temperature
      optimization.
    kl_constraints: KL and variables for applying Lagrangian penalties to bound
      them in the M-step, KLs are E* or [E*, A]. Here A is the action dimension
      in the case of per-dimension KL constraints.
    projection_operator: Function to project dual variables (temperature and kl
      constraint alphas) into the positive range.
    restarting_weights: Optional restarting weights, shape E*, 0 means that this
      step is the start of a new episode and we ignore losses at this step
      because the agent cannot influence these.
    importance_weights: Optional importance weights, shape E*.
    top_k_fraction: Fraction of samples to use in the E-step.
    policy_loss_weight: Weight for the policy loss.
    temperature_loss_weight: Weight for the temperature loss.
    kl_loss_weight: Weight for the KL loss.
    alpha_loss_weight: Weight for the alpha loss.
    axis_name: Optional axis name for `pmap`. If `None`, computations
      are performed locally on each device.
    use_stop_gradient: bool indicating whether or not to apply stop gradient.

  Returns:
    Per example `loss` with same shape E* as array inputs, and additional data
    including the components of this loss and the normalized weights in the
    AdditionalOutputs.
  """
  # Define default restarting weights and importance weights.
  if restarting_weights is None:
    restarting_weights = jnp.ones_like(sample_log_probs)
  if importance_weights is None:
    importance_weights = jnp.ones_like(sample_log_probs)

  # Check shapes.
  chex.assert_equal_shape(
      [advantages, sample_log_probs, restarting_weights, importance_weights])

  chex.assert_rank(temperature_constraint.epsilon, 0)
  chex.assert_type([
      sample_log_probs, advantages, restarting_weights, importance_weights,
      temperature_constraint.alpha, temperature_constraint.epsilon], float)

  for kl, penalty in kl_constraints:
    chex.assert_rank(penalty.epsilon, 0)
    chex.assert_type([kl, penalty.alpha, penalty.epsilon], float)
    if penalty.per_dimension:
      chex.assert_rank(kl, advantages.ndim + 1)
      chex.assert_equal_shape_prefix([kl, advantages], advantages.ndim)
    else:
      chex.assert_equal_shape([kl, advantages])

  # E-step: Calculate the reweighting and the temperature loss.
  temperature_loss, norm_weights, num_samples = (
      vmpo_compute_weights_and_temperature_loss(
          advantages, restarting_weights, importance_weights,
          temperature_constraint, projection_operator, top_k_fraction,
          axis_name=axis_name, use_stop_gradient=use_stop_gradient))

  # M-step: Supervised learning of reweighted trajectories using the weights
  # from the E-step, with additional KL constraints.
  # The weights are normalized so that the sum is 1. We multiply by the number
  # of examples so that we can give a policy loss per example and take the mean,
  # and we assume `restarting_weights` are already included.
  if axis_name:
    num_examples = jax.lax.all_gather(
        sample_log_probs, axis_name=axis_name).size
  else:
    num_examples = sample_log_probs.size
  policy_loss = -sample_log_probs * norm_weights * num_examples

  kl_loss, alpha_loss = compute_parametric_kl_penalty_and_dual_loss(
      kl_constraints, projection_operator, use_stop_gradient)

  chex.assert_equal_shape([policy_loss, kl_loss, alpha_loss])

  # Calculate the total policy improvement loss.
  loss = (policy_loss_weight * policy_loss +
          temperature_loss_weight * temperature_loss +
          kl_loss_weight * kl_loss +
          alpha_loss_weight * alpha_loss)

  return loss, MpoOutputs(
      temperature_loss=temperature_loss, policy_loss=policy_loss,
      kl_loss=kl_loss, alpha_loss=alpha_loss, normalized_weights=norm_weights,
      num_samples=num_samples)


def get_top_k_weights(
    top_k_fraction: float,
    restarting_weights: Array,
    scaled_advantages: Array,
    axis_name: Optional[str] = None,
    use_stop_gradient: bool = True,
):
  """Get the weights for the top top_k_fraction of advantages.

  Args:
    top_k_fraction: The fraction of weights to use.
    restarting_weights: Restarting weights, shape E*, 0 means that this step is
      the start of a new episode and we ignore losses at this step because the
      agent cannot influence these.
    scaled_advantages: The advantages for each example (shape E*), scaled by
      temperature.
    axis_name: Optional axis name for `pmap`. If `None`, computations are
      performed locally on each device.
    use_stop_gradient: bool indicating whether or not to apply stop gradient.

  Returns:
    Weights for the top top_k_fraction of advantages
  """
  chex.assert_equal_shape([scaled_advantages, restarting_weights])
  chex.assert_type([scaled_advantages, restarting_weights], float)

  if not 0.0 < top_k_fraction <= 1.0:
    raise ValueError(
        f"`top_k_fraction` must be in (0, 1], got {top_k_fraction}")
  logging.info("[vmpo_e_step] top_k_fraction: %f", top_k_fraction)

  if top_k_fraction < 1.0:
    # Don't include the restarting samples in the determination of top-k.
    valid_scaled_advantages = scaled_advantages - (
        1.0 - restarting_weights) * _INFINITY
    # Determine the minimum top-k value across all devices,
    if axis_name:
      all_valid_scaled_advantages = jax.lax.all_gather(
          valid_scaled_advantages, axis_name=axis_name)
    else:
      all_valid_scaled_advantages = valid_scaled_advantages
    top_k = int(top_k_fraction * jnp.size(all_valid_scaled_advantages))
    if top_k == 0:
      raise ValueError(
          "top_k_fraction too low to get any valid scaled advantages.")
    # TODO(b/160450251): Use jnp.partition(all_valid_scaled_advantages, top_k)
    #   when this is implemented in jax.
    top_k_min = jnp.sort(jnp.reshape(all_valid_scaled_advantages, [-1]))[-top_k]
    # Fold the top-k into the restarting weights.
    top_k_weights = jnp.greater_equal(valid_scaled_advantages,
                                      top_k_min).astype(jnp.float32)
    top_k_weights = jax.lax.select(
        use_stop_gradient, jax.lax.stop_gradient(top_k_weights), top_k_weights)
    top_k_restarting_weights = restarting_weights * top_k_weights
  else:
    top_k_restarting_weights = restarting_weights

  return top_k_restarting_weights


def vmpo_compute_weights_and_temperature_loss(
    advantages: Array,
    restarting_weights: Array,
    importance_weights: Array,
    temperature_constraint: LagrangePenalty,
    projection_operator: Callable[[Numeric], Numeric],
    top_k_fraction: float,
    axis_name: Optional[str] = None,
    use_stop_gradient: bool = True,
) -> Tuple[Scalar, Array, Scalar]:
  """Computes the weights and temperature loss for V-MPO.

  Args:
    advantages: Advantages for the E-step. Shape E*.
    restarting_weights: Restarting weights, 0 means that this
      step is the start of a new episode and we ignore losses at this step
      because the agent cannot influence these. Shape E*.
    importance_weights: Optional importance weights. Shape E*
    temperature_constraint: Lagrange constraint for the E-step temperature
      optimization.
    projection_operator: Function to project dual variables (temperature and kl
      constraint alphas) into the positive range.
    top_k_fraction: Fraction of samples to use in the E-step.
    axis_name: Optional axis name for `pmap` or 'vmap'. If `None`, computations
      are performed locally on each device.
    use_stop_gradient: bool indicating whether or not to apply stop gradient.

  Returns:
    The temperature loss, normalized weights and number of samples used.
  """
  chex.assert_equal_shape([advantages, restarting_weights, importance_weights])
  chex.assert_rank(temperature_constraint.epsilon, 0)
  chex.assert_type([
      advantages, restarting_weights, importance_weights,
      temperature_constraint.alpha, temperature_constraint.epsilon], float)

  importance_weights = jax.lax.select(
      use_stop_gradient, jax.lax.stop_gradient(importance_weights),
      importance_weights)

  # Lagrange constraint.
  temperature = projection_operator(temperature_constraint.alpha)
  epsilon_temperature = temperature_constraint.epsilon

  # Scale the advantages.
  scaled_advantages = restarting_weights * advantages / temperature
  max_scaled_advantage = jnp.max(scaled_advantages)
  # If the axis_name is not None find the maximum across all devices.
  if axis_name:
    assert use_stop_gradient  # Cant differentiate through pmax.
    max_scaled_advantage = jax.lax.stop_gradient(max_scaled_advantage)
    max_scaled_advantage = jax.lax.pmax(
        max_scaled_advantage, axis_name=axis_name)
  else:
    max_scaled_advantage = jax.lax.select(
        use_stop_gradient, jax.lax.stop_gradient(max_scaled_advantage),
        max_scaled_advantage)
  # Maybe don't use all of the advantages.
  top_k_restarting_weights = get_top_k_weights(
      top_k_fraction, restarting_weights, scaled_advantages, axis_name,
      use_stop_gradient)

  all_sum = base.AllSum(axis_name)

  # Reweight the old trajectories.
  unnormalized_weights = (top_k_restarting_weights * importance_weights
                          * jnp.exp(scaled_advantages - max_scaled_advantage))
  # If the axis_name is not None these sums will be taken across all devices.
  sum_weights = all_sum(unnormalized_weights) + _EPSILON
  num_samples = all_sum(top_k_restarting_weights) + _EPSILON

  normalized_weights = unnormalized_weights / sum_weights
  normalized_weights = jax.lax.select(use_stop_gradient,
                                      jax.lax.stop_gradient(normalized_weights),
                                      normalized_weights)

  # Calculate the temperature loss.
  log_mean_weights = (jnp.log(sum_weights) + max_scaled_advantage
                      - jnp.log(num_samples))
  temperature_loss = temperature * (epsilon_temperature + log_mean_weights)

  return temperature_loss, normalized_weights, num_samples


def kl_constraint_loss(
    kl: Array,
    penalty: LagrangePenalty,
    projection_operator: Callable[[Numeric], Numeric],
    use_stop_gradient: bool = True,
) -> Tuple[Array, Array, Array]:
  """Implements a hard KL constraint.

  The optimization proceeds in two phases. First, we optimize the weighting
  term `alpha` keeping the KL constant and then we optimize the KL keeping
  `alpha` constant. Each phase is implemented by appropriately using a
  stop_gradient.

  If `bound` - `kl` > 0, then `alpha` is pushed toward `min_alpha`. However
  this also means the policy is free to change more since `kl` < `bound `.
  This eventually leads to `bound` - `kl` < 0 which pressures alpha to get to
  a high positive value. This coordinate ascent results in `kl` staying close
  to `bound`.

  Args:
    kl: The kl per example which is being constrained.
    penalty: The dual variable used to impose a penalty and parameters of the
      constraint.
    projection_operator: Function to project dual variables kl constraint alphas
      into the positive range.
    use_stop_gradient: bool indicating whether or not to apply stop gradient.

  Returns:
    A `tuple` consisting of three arrays: `kl_loss`, `alpha_loss` and
    `clipped_alpha`. The first two terms represent the losses for the two phases
    of computation. `clipped_alpha` is the clipped lagrangian multiplier `alpha`
    that is learnt.
  """
  chex.assert_type([kl, penalty.alpha, penalty.epsilon], float)

  alpha = projection_operator(penalty.alpha)
  alpha_constant = jax.lax.select(
      use_stop_gradient, jax.lax.stop_gradient(penalty.alpha), penalty.alpha)

  # First step: Optimize w.r.t. alphas
  alpha_loss = alpha * (
      penalty.epsilon -
      jax.lax.select(use_stop_gradient, jax.lax.stop_gradient(kl), kl))

  # Second step: KL loss.
  kl_loss = alpha_constant * kl

  # If kl_loss and alpha_loss are per dimension then at this point sum over
  # dimensions.
  if penalty.per_dimension:
    kl_loss = jnp.sum(kl_loss, axis=-1)
    alpha_loss = jnp.sum(alpha_loss, axis=-1)

  return kl_loss, alpha_loss, alpha


def kl_alpha_loss(
    restarting_weights: Array,
    kl_constraints: Sequence[Tuple[Array, LagrangePenalty]] = (),
    axis_name: Optional[str] = None):
  """Calculates the losses for multiple KL constraints.

  Args:
    restarting_weights: Restarting weights, shape E*, 0 means that this step is
      the start of a new episode and we ignore losses at this step because the
      agent cannot influence these.
    kl_constraints: KL and variables for applying Lagrangian penalties to bound
      them in the M-step, KLs are [E*, A?]. Here A is the action dimension
      in the case of per-dimension KL constraints.
    axis_name: Optional axis name for `pmap`. If `None`, computations are
      performed locally on each device.

  Returns:
    The kl loss and dual variable loss both shape E*.
  """
  chex.assert_type(restarting_weights, float)
  if kl_constraints:
    for kl, penalty in kl_constraints:
      chex.assert_rank(penalty.epsilon, 0)
      chex.assert_type([kl, penalty.alpha, penalty.epsilon], float)
      chex.assert_equal_shape_prefix([kl, restarting_weights],
                                     restarting_weights.ndim)

    # Implement decoupled KL constraints.
    kl_alpha_losses = [kl_constraint_loss(kl, penalty, lambda x: x)[:2]
                       for kl, penalty in kl_constraints]
    kl_loss, alpha_loss = [sum(losses) for losses in zip(*kl_alpha_losses)]
    all_sum = base.AllSum(axis_name)
    num_samples = all_sum(restarting_weights) + _EPSILON
    # Reshape in case KL is per dimension.
    kl_loss = all_sum(kl_loss * restarting_weights) / num_samples
    alpha_loss = all_sum(alpha_loss * restarting_weights) / num_samples
  else:
    # No M-step constraint.
    kl_loss = jnp.asarray(0.0)
    alpha_loss = jnp.asarray(0.0)
  return kl_loss, alpha_loss




# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions for calculating multistep returns.

In this subpackage we expose a number of functions that may be used to compute
multistep truncated bootstrapped estimates of the return (the discounted sum of
rewards collected by an agent). These estimate compute returns from trajectories
of experience; trajectories are not assumed to align with episode boundaries,
and bootstrapping is used to estimate returns beyond the end of a trajectory.
"""
from typing import Union
import chex
import jax
import jax.numpy as jnp
from rlax._src import base

Array = chex.Array
Scalar = chex.Scalar
Numeric = chex.Numeric


def lambda_returns(
    r_t: Array,
    discount_t: Array,
    v_t: Array,
    lambda_: Numeric = 1.,
    stop_target_gradients: bool = False,
) -> Array:
  """Estimates a multistep truncated lambda return from a trajectory.

  Given a a trajectory of length `T+1`, generated under some policy π, for each
  time-step `t` we can estimate a target return `G_t`, by combining rewards,
  discounts, and state values, according to a mixing parameter `lambda`.

  The parameter `lambda_`  mixes the different multi-step bootstrapped returns,
  corresponding to accumulating `k` rewards and then bootstrapping using `v_t`.

    rₜ₊₁ + γₜ₊₁ vₜ₊₁
    rₜ₊₁ + γₜ₊₁ rₜ₊₂ + γₜ₊₁ γₜ₊₂ vₜ₊₂
    rₜ₊₁ + γₜ₊₁ rₜ₊₂ + γₜ₊₁ γₜ₊₂ rₜ₊₂ + γₜ₊₁ γₜ₊₂ γₜ₊₃ vₜ₊₃

  The returns are computed recursively, from `G_{T-1}` to `G_0`, according to:

    Gₜ = rₜ₊₁ + γₜ₊₁ [(1 - λₜ₊₁) vₜ₊₁ + λₜ₊₁ Gₜ₊₁].

  In the `on-policy` case, we estimate a return target `G_t` for the same
  policy π that was used to generate the trajectory. In this setting the
  parameter `lambda_` is typically a fixed scalar factor. Depending
  on how values `v_t` are computed, this function can be used to construct
  targets for different multistep reinforcement learning updates:

    TD(λ):  `v_t` contains the state value estimates for each state under π.
    Q(λ):  `v_t = max(q_t, axis=-1)`, where `q_t` estimates the action values.
    Sarsa(λ):  `v_t = q_t[..., a_t]`, where `q_t` estimates the action values.

  In the `off-policy` case, the mixing factor is a function of state, and
  different definitions of `lambda` implement different off-policy corrections:

    Per-decision importance sampling:  λₜ = λ ρₜ = λ [π(aₜ|sₜ) / μ(aₜ|sₜ)]
    V-trace, as instantiated in IMPALA:  λₜ = min(1, ρₜ)

  Note that the second option is equivalent to applying per-decision importance
  sampling, but using an adaptive λ(ρₜ) = min(1/ρₜ, 1), such that the effective
  bootstrap parameter at time t becomes λₜ = λ(ρₜ) * ρₜ = min(1, ρₜ).
  This is the interpretation used in the ABQ(ζ) algorithm (Mahmood 2017).

  Of course this can be augmented to include an additional factor λ.  For
  instance we could use V-trace with a fixed additional parameter λ = 0.9, by
  setting λₜ = 0.9 * min(1, ρₜ) or, alternatively (but not equivalently),
  λₜ = min(0.9, ρₜ).

  Estimated return are then often used to define a td error, e.g.:  ρₜ(Gₜ - vₜ).

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/sutton/book/ebook/node74.html).

  Args:
    r_t: sequence of rewards rₜ for timesteps t in [1, T].
    discount_t: sequence of discounts γₜ for timesteps t in [1, T].
    v_t: sequence of state values estimates under π for timesteps t in [1, T].
    lambda_: mixing parameter; a scalar or a vector for timesteps t in [1, T].
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Multistep lambda returns.
  """
  chex.assert_rank([r_t, discount_t, v_t, lambda_], [1, 1, 1, {0, 1}])
  chex.assert_type([r_t, discount_t, v_t, lambda_], float)
  chex.assert_equal_shape([r_t, discount_t, v_t])

  # If scalar make into vector.
  lambda_ = jnp.ones_like(discount_t) * lambda_

  # Work backwards to compute `G_{T-1}`, ..., `G_0`.
  def _body(acc, xs):
    returns, discounts, values, lambda_ = xs
    acc = returns + discounts * ((1-lambda_) * values + lambda_ * acc)
    return acc, acc

  _, returns = jax.lax.scan(
      _body, v_t[-1], (r_t, discount_t, v_t, lambda_), reverse=True)

  return jax.lax.select(stop_target_gradients,
                        jax.lax.stop_gradient(returns),
                        returns)


def n_step_bootstrapped_returns(
    r_t: Array,
    discount_t: Array,
    v_t: Array,
    n: int,
    lambda_t: Numeric = 1.,
    stop_target_gradients: bool = False,
) -> Array:
  """Computes strided n-step bootstrapped return targets over a sequence.

  The returns are computed according to the below equation iterated `n` times:

     Gₜ = rₜ₊₁ + γₜ₊₁ [(1 - λₜ₊₁) vₜ₊₁ + λₜ₊₁ Gₜ₊₁].

  When lambda_t == 1. (default), this reduces to

     Gₜ = rₜ₊₁ + γₜ₊₁ * (rₜ₊₂ + γₜ₊₂ * (... * (rₜ₊ₙ + γₜ₊ₙ * vₜ₊ₙ ))).

  Args:
    r_t: rewards at times [1, ..., T].
    discount_t: discounts at times [1, ..., T].
    v_t: state or state-action values to bootstrap from at time [1, ...., T].
    n: number of steps over which to accumulate reward before bootstrapping.
    lambda_t: lambdas at times [1, ..., T]. Shape is [], or [T-1].
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    estimated bootstrapped returns at times [0, ...., T-1]
  """
  chex.assert_rank([r_t, discount_t, v_t, lambda_t], [1, 1, 1, {0, 1}])
  chex.assert_type([r_t, discount_t, v_t, lambda_t], float)
  chex.assert_equal_shape([r_t, discount_t, v_t])
  seq_len = r_t.shape[0]

  # Maybe change scalar lambda to an array.
  lambda_t = jnp.ones_like(discount_t) * lambda_t

  # Shift bootstrap values by n and pad end of sequence with last value v_t[-1].
  pad_size = min(n - 1, seq_len)
  targets = jnp.concatenate([v_t[n - 1:], jnp.array([v_t[-1]] * pad_size)])

  # Pad sequences. Shape is now (T + n - 1,).
  r_t = jnp.concatenate([r_t, jnp.zeros(n - 1)])
  discount_t = jnp.concatenate([discount_t, jnp.ones(n - 1)])
  lambda_t = jnp.concatenate([lambda_t, jnp.ones(n - 1)])
  v_t = jnp.concatenate([v_t, jnp.array([v_t[-1]] * (n - 1))])

  # Work backwards to compute n-step returns.
  for i in reversed(range(n)):
    r_ = r_t[i:i + seq_len]
    discount_ = discount_t[i:i + seq_len]
    lambda_ = lambda_t[i:i + seq_len]
    v_ = v_t[i:i + seq_len]
    targets = r_ + discount_ * ((1. - lambda_) * v_ + lambda_ * targets)

  return jax.lax.select(stop_target_gradients,
                        jax.lax.stop_gradient(targets), targets)


def discounted_returns(
    r_t: Array,
    discount_t: Array,
    v_t: Array,
    stop_target_gradients: bool = False,
) -> Array:
  """Calculates a discounted return from a trajectory.

  The returns are computed recursively, from `G_{T-1}` to `G_0`, according to:

    Gₜ = rₜ₊₁ + γₜ₊₁ Gₜ₊₁.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/sutton/book/ebook/node61.html).

  Args:
    r_t: reward sequence at time t.
    discount_t: discount sequence at time t.
    v_t: value sequence or scalar at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Discounted returns.
  """
  chex.assert_rank([r_t, discount_t, v_t], [1, 1, {0, 1}])
  chex.assert_type([r_t, discount_t, v_t], float)

  # If scalar make into vector.
  bootstrapped_v = jnp.ones_like(discount_t) * v_t
  return lambda_returns(r_t, discount_t, bootstrapped_v, lambda_=1.,
                        stop_target_gradients=stop_target_gradients)


def importance_corrected_td_errors(
    r_t: Array,
    discount_t: Array,
    rho_tm1: Array,
    lambda_: Array,
    values: Array,
    stop_target_gradients: bool = False,
) -> Array:
  """Computes the multistep td errors with per decision importance sampling.

  Given a trajectory of length `T+1`, generated under some policy π, for each
  time-step `t` we can estimate a multistep temporal difference error δₜ(ρ,λ),
  by combining rewards, discounts, and state values, according to a mixing
  parameter `λ` and importance sampling ratios ρₜ = π(aₜ|sₜ) / μ(aₜ|sₜ):

    td-errorₜ = ρₜ δₜ(ρ,λ)
    δₜ(ρ,λ) = δₜ + ρₜ₊₁ λₜ₊₁ γₜ₊₁ δₜ₊₁(ρ,λ),

  where δₜ = rₜ₊₁ + γₜ₊₁ vₜ₊₁ - vₜ is the one step, temporal difference error
  for the agent's state value estimates. This is equivalent to computing
  the λ-return with λₜ = ρₜ (e.g. using the `lambda_returns` function from
  above), and then computing errors as  td-errorₜ = ρₜ(Gₜ - vₜ).

  See "A new Q(λ) with interim forward view and Monte Carlo equivalence"
  by Sutton et al. (http://proceedings.mlr.press/v32/sutton14.html).

  Args:
    r_t: sequence of rewards rₜ for timesteps t in [1, T].
    discount_t: sequence of discounts γₜ for timesteps t in [1, T].
    rho_tm1: sequence of importance ratios for all timesteps t in [0, T-1].
    lambda_: mixing parameter; scalar or have per timestep values in [1, T].
    values: sequence of state values under π for all timesteps t in [0, T].
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Off-policy estimates of the multistep td errors.
  """
  chex.assert_rank([r_t, discount_t, rho_tm1, values], [1, 1, 1, 1])
  chex.assert_type([r_t, discount_t, rho_tm1, values], float)
  chex.assert_equal_shape([r_t, discount_t, rho_tm1, values[1:]])

  v_tm1 = values[:-1]  # Predictions to compute errors for.
  v_t = values[1:]  # Values for bootstrapping.
  rho_t = jnp.concatenate((rho_tm1[1:], jnp.array([1.])))  # Unused dummy value.
  lambda_ = jnp.ones_like(discount_t) * lambda_  # If scalar, make into vector.

  # Compute the one step temporal difference errors.
  one_step_delta = r_t + discount_t * v_t - v_tm1

  # Work backwards to compute `delta_{T-1}`, ..., `delta_0`.
  def _body(acc, xs):
    deltas, discounts, rho_t, lambda_ = xs
    acc = deltas + discounts * rho_t * lambda_ * acc
    return acc, acc

  _, errors = jax.lax.scan(
      _body, 0.0, (one_step_delta, discount_t, rho_t, lambda_), reverse=True)

  errors = rho_tm1 * errors
  return jax.lax.select(stop_target_gradients,
                        jax.lax.stop_gradient(errors + v_tm1) - v_tm1, errors)


def truncated_generalized_advantage_estimation(
    r_t: Array,
    discount_t: Array,
    lambda_: Union[Array, Scalar],
    values: Array,
    stop_target_gradients: bool = False,
) -> Array:
  """Computes truncated generalized advantage estimates for a sequence length k.

  The advantages are computed in a backwards fashion according to the equation:
  Âₜ = δₜ + (γλ) * δₜ₊₁ + ... + ... + (γλ)ᵏ⁻ᵗ⁺¹ * δₖ₋₁
  where δₜ = rₜ₊₁ + γₜ₊₁ * v(sₜ₊₁) - v(sₜ).

  See Proximal Policy Optimization Algorithms, Schulman et al.:
  https://arxiv.org/abs/1707.06347

  Note: This paper uses a different notation than the RLax standard
  convention that follows Sutton & Barto. We use rₜ₊₁ to denote the reward
  received after acting in state sₜ, while the PPO paper uses rₜ.

  Args:
    r_t: Sequence of rewards at times [1, k]
    discount_t: Sequence of discounts at times [1, k]
    lambda_: Mixing parameter; a scalar or sequence of lambda_t at times [1, k]
    values: Sequence of values under π at times [0, k]
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Multistep truncated generalized advantage estimation at times [0, k-1].
  """
  chex.assert_rank([r_t, values, discount_t], 1)
  chex.assert_type([r_t, values, discount_t], float)
  lambda_ = jnp.ones_like(discount_t) * lambda_  # If scalar, make into vector.

  delta_t = r_t + discount_t * values[1:] - values[:-1]

  # Iterate backwards to calculate advantages.
  def _body(acc, xs):
    deltas, discounts, lambda_ = xs
    acc = deltas + discounts * lambda_ * acc
    return acc, acc

  _, advantage_t = jax.lax.scan(
      _body, 0.0, (delta_t, discount_t, lambda_), reverse=True)

  return jax.lax.select(stop_target_gradients,
                        jax.lax.stop_gradient(advantage_t),
                        advantage_t)


def general_off_policy_returns_from_action_values(
    q_t: Array,
    a_t: Array,
    r_t: Array,
    discount_t: Array,
    c_t: Array,
    pi_t: Array,
    stop_target_gradients: bool = False,
) -> Array:
  """Calculates targets for various off-policy correction algorithms.

  Given a window of experience of length `K`, generated by a behaviour policy μ,
  for each time-step `t` we can estimate the return `G_t` from that step
  onwards, under some target policy π, using the rewards in the trajectory, the
  actions selected by μ and the action-values under π, according to equation:

    Gₜ = rₜ₊₁ + γₜ₊₁ * (E[q(aₜ₊₁)] - cₜ * q(aₜ₊₁) + cₜ * Gₜ₊₁),

  where, depending on the choice of `c_t`, the algorithm implements:

    Importance Sampling             c_t = π(x_t, a_t) / μ(x_t, a_t),
    Harutyunyan's et al. Q(lambda)  c_t = λ,
    Precup's et al. Tree-Backup     c_t = π(x_t, a_t),
    Munos' et al. Retrace           c_t = λ min(1, π(x_t, a_t) / μ(x_t, a_t)).

  See "Safe and Efficient Off-Policy Reinforcement Learning" by Munos et al.
  (https://arxiv.org/abs/1606.02647).

  Args:
    q_t: Q-values at times [1, ..., K - 1].
    a_t: action index at times [1, ..., K - 1].
    r_t: reward at times [1, ..., K - 1].
    discount_t: discount at times [1, ..., K - 1].
    c_t: importance weights at times [1, ..., K - 1].
    pi_t: target policy probs at times [1, ..., K - 1].
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Off-policy estimates of the generalized returns from states visited at times
    [0, ..., K - 1].
  """
  chex.assert_rank([q_t, a_t, r_t, discount_t, c_t, pi_t], [2, 1, 1, 1, 1, 2])
  chex.assert_type([q_t, a_t, r_t, discount_t, c_t, pi_t],
                   [float, int, float, float, float, float])
  chex.assert_equal_shape(
      [q_t[..., 0], a_t, r_t, discount_t, c_t, pi_t[..., 0]])

  # Get the expected values and the values of actually selected actions.
  exp_q_t = (pi_t * q_t).sum(axis=-1)
  # The generalized returns are independent of Q-values and cs at the final
  # state.
  q_a_t = base.batched_index(q_t, a_t)[:-1]
  c_t = c_t[:-1]

  return general_off_policy_returns_from_q_and_v(
      q_a_t, exp_q_t, r_t, discount_t, c_t, stop_target_gradients)


def general_off_policy_returns_from_q_and_v(
    q_t: Array,
    v_t: Array,
    r_t: Array,
    discount_t: Array,
    c_t: Array,
    stop_target_gradients: bool = False,
) -> Array:
  """Calculates targets for various off-policy evaluation algorithms.

  Given a window of experience of length `K+1`, generated by a behaviour policy
  μ, for each time-step `t` we can estimate the return `G_t` from that step
  onwards, under some target policy π, using the rewards in the trajectory, the
  values under π of states and actions selected by μ, according to equation:

    Gₜ = rₜ₊₁ + γₜ₊₁ * (vₜ₊₁ - cₜ₊₁ * q(aₜ₊₁) + cₜ₊₁* Gₜ₊₁),

  where, depending on the choice of `c_t`, the algorithm implements:

    Importance Sampling             c_t = π(x_t, a_t) / μ(x_t, a_t),
    Harutyunyan's et al. Q(lambda)  c_t = λ,
    Precup's et al. Tree-Backup     c_t = π(x_t, a_t),
    Munos' et al. Retrace           c_t = λ min(1, π(x_t, a_t) / μ(x_t, a_t)).

  See "Safe and Efficient Off-Policy Reinforcement Learning" by Munos et al.
  (https://arxiv.org/abs/1606.02647).

  Args:
    q_t: Q-values under π of actions executed by μ at times [1, ..., K - 1].
    v_t: Values under π at times [1, ..., K].
    r_t: rewards at times [1, ..., K].
    discount_t: discounts at times [1, ..., K].
    c_t: weights at times [1, ..., K - 1].
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Off-policy estimates of the generalized returns from states visited at times
    [0, ..., K - 1].
  """
  chex.assert_rank([q_t, v_t, r_t, discount_t, c_t], 1)
  chex.assert_type([q_t, v_t, r_t, discount_t, c_t], float)
  chex.assert_equal_shape([q_t, v_t[:-1], r_t[:-1], discount_t[:-1], c_t])

  g = r_t[-1] + discount_t[-1] * v_t[-1]  # G_K-1.

  def _body(acc, xs):
    reward, discount, c, v, q = xs
    acc = reward + discount * (v - c * q + c * acc)
    return acc, acc

  _, returns = jax.lax.scan(
      _body, g, (r_t[:-1], discount_t[:-1], c_t, v_t[:-1], q_t), reverse=True)
  returns = jnp.concatenate([returns, g[jnp.newaxis]], axis=0)

  return jax.lax.select(stop_target_gradients,
                        jax.lax.stop_gradient(returns),
                        returns)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions implementing PopArt."""

import collections

from typing import Mapping, Tuple

import chex
import jax
import jax.numpy as jnp
from rlax._src import base

Array = chex.Array

LinearParams = Mapping[str, Array]
PopArtState = collections.namedtuple(
    "PopArtState", ["shift", "scale", "second_moment"])
PopArtOutput = collections.namedtuple(
    "PopArtOutput", ["normalized", "unnormalized"])


def _cross_replica_scatter_add(source: Array, indices: Array, updates: Array,
                               axis_name):
  """tf.scatter_add, but with JAX, cross replica, and without state.

  Args:
    source: An array of shape [O].
    indices: An array indicating which index each update is for.
    updates: The updates to apply to `source`. Of same shape as indices.
    axis_name: What axis to aggregate over, if str. If passed an iterable,
      aggregates over multiple axes. Defaults to no aggregation, i.e. None.

  Returns:
    An array of shape [O], which is source + the scattered updates from all
    replicas.
  """
  assert updates.shape == indices.shape
  assert jnp.issubdtype(indices.dtype, jnp.integer)
  assert source.ndim == 1
  # Flatten indices, updates.
  num_classes = source.shape[0]
  indices = jnp.reshape(indices, [-1])
  updates = jnp.reshape(updates, [-1])
  # Scatter updates according to value of indices.
  updates_at_idxs = updates[..., None] * base.one_hot(indices, num_classes)
  # Aggregate locally first, then across replicas.
  total_updates = jnp.sum(updates_at_idxs, axis=0)
  if axis_name is not None:
    axis_names = (axis_name,) if isinstance(axis_name, str) else axis_name
    for a_name in axis_names:
      total_updates = jax.lax.psum(total_updates, axis_name=a_name)
  return source + total_updates


def normalize(state: PopArtState, unnormalized: Array, indices: Array) -> Array:
  """Returns normalized values.

  Args:
    state: The PopArt summary stats.
    unnormalized: unnormalized values that we applied PopArt to.
    indices: Which scale and shifts to use

  Returns:
    Normalized PopArt values.
  """
  scale = state.scale[indices]
  shift = state.shift[indices]
  normalized = (unnormalized - shift) / scale
  return normalized


def unnormalize(state: PopArtState, normalized: Array, indices: Array) -> Array:
  """Returns unnormalized values.

  Args:
    state: The PopArt summary stats.
    normalized: normalized values that we apply PopArt to.
    indices: Which scale and shifts to use

  Returns:
    Unnormalized PopArt values.
  """
  scale = state.scale[indices]
  shift = state.shift[indices]
  unnormalized = scale * normalized + shift
  return unnormalized


def unnormalize_linear(state: PopArtState, inputs: Array,
                       indices: Array) -> PopArtOutput:
  """Selects and unnormalizes output of a Linear.

  Args:
    state: The PopArt summary stats.
    inputs: The (normalized) output of the Linear that we apply PopArt to.
    indices: Which indices of `inputs` to use.

  Returns:
    PopArtOutput, a tuple of the normalized and unnormalized PopArt values.
  """
  assert jnp.issubdtype(indices.dtype, jnp.integer)
  assert indices.shape == inputs.shape[:-1]
  normalized = jnp.take_along_axis(inputs, indices[..., None], axis=-1)
  normalized = jnp.squeeze(normalized, axis=-1)
  return PopArtOutput(normalized, unnormalize(state, normalized, indices))


def art(state: PopArtState,
        targets: Array,
        indices: Array,
        step_size: float,
        scale_lb: float,
        scale_ub: float,
        axis_name=None) -> PopArtState:
  """Adaptively rescale targets.

  Args:
    state: The PopArt summary stats.
    targets: targets which are rescaled.
    indices: Which indices of the state to use.
    step_size: The step size for learning the scale & shift parameters.
    scale_lb: Lower bound for the scale.
    scale_ub: Upper bound for the scale.
    axis_name: What axis to aggregate over, if str. If passed an iterable,
      aggregates over multiple axes. Defaults to no aggregation, i.e. None.

  Returns:
    New popart state which can be used to rescale targets.
  """
  assert targets.shape == indices.shape
  assert jnp.issubdtype(indices.dtype, jnp.integer)

  # Update shift.
  shift_gather = state.shift[indices]
  shift_update = step_size * (targets - shift_gather)
  shift_new = _cross_replica_scatter_add(state.shift, indices, shift_update,
                                         axis_name)

  # Update second moment.
  second_moment_gather = state.second_moment[indices]
  second_moment_update = step_size * (
      jnp.square(targets) - second_moment_gather)
  second_moment_new = _cross_replica_scatter_add(state.second_moment, indices,
                                                 second_moment_update,
                                                 axis_name)

  # Derive scale (stdev) from second moment and mean.
  scale_sq = second_moment_new - jnp.square(shift_new)
  scale_sq = jnp.clip(scale_sq, scale_lb**2, scale_ub**2)
  scale_new = jnp.sqrt(scale_sq)

  state_new = PopArtState(shift_new, scale_new, second_moment_new)

  # Prevent gradients propagating back through the state.
  state_new = jax.tree_map(jax.lax.stop_gradient, state_new)

  return state_new


def pop(params: LinearParams, old: PopArtState, new: PopArtState):
  """Preserves outputs precisely.

  Args:
    params: The parameters of the linear to preserve.
    old: The old PopArt state.
    new: The new PopArt state.

  Returns:
    new parameters.
  """
  w_new = params["w"] * jnp.broadcast_to(old.scale / new.scale,
                                         params["w"].shape)
  b_new = (old.scale * params["b"] + old.shift - new.shift) / new.scale
  params_new = dict(w=w_new, b=b_new)
  return params_new


def popart(num_outputs: int,
           step_size: float,
           scale_lb: float,
           scale_ub: float,
           axis_name=None):
  """Generates functions giving initial PopArt state and update rule.

  Args:
    num_outputs: The number of outputs generated by the linear we're preserving.
    step_size: The step size for learning the scale & shift parameters.
    scale_lb: Lower bound for the scale.
    scale_ub: Upper bound for the scale.
    axis_name: What axis to aggregate over, if str. If passed an iterable,
      aggregates over multiple axes. Defaults to no aggregation, i.e. None.

  Returns:
    A tuple of:
      initial_state: A function returning the initial PopArt state.
      popart_update: A function updating the PopArt state and parameters
      of the preceding linear.
  """

  def initial_state():
    return PopArtState(
        jnp.zeros([num_outputs]), jnp.ones([num_outputs]),
        jnp.ones([num_outputs]))

  def popart_update(params: LinearParams, state: PopArtState, targets: Array,
                    indices: Array) -> Tuple[LinearParams, PopArtState]:
    """Computes the PopArt update.

    Args:
      params: The parameters of the linear to preserve.
      state: The current PopArt state.
      targets: Values whose distribution to learn.
      indices: For each target, which shift and scale element to adjust.

    Returns:
      A tuple of:
        new_params: The new parameters of the linear, preserving outputs.
        new_state: The new PopArt state.
    """
    # Disables Popart if step_size is None
    if step_size is None:
      return params, state

    # Adaptively rescale targets.
    state_new = art(state, targets, indices, step_size, scale_lb, scale_ub,
                    axis_name)
    # Preserve outputs precisely.
    params_new = pop(params, state, state_new)
    return params_new, state_new

  return initial_state, popart_update
